"""Wrapper objects for Unreal code templates."""

# Standard Library
import copy
import re
from typing import Set


class CodeTemplateError(Exception):
    """Custom exception representing errors encountered with CodeTemplate."""


class CodeTemplate(object):
    TOKEN_RE = re.compile(r"\%([A-Z_]+)\%", flags=re.MULTILINE)

    def __init__(self, name: str, template_string: str = "") -> None:
        self.name: str = name
        self.template_string: str = template_string

    def __repr__(self) -> str:
        """Python interpreter representation of CodeTemplate."""
        return f"<CodeTemplate {self.name}>"

    @property
    def tokens(self) -> Set[str]:
        return {
            token
            for match in CodeTemplate.TOKEN_RE.finditer(self.template_string)
            for token in match.groups()
        }

    def make_instance(self, **tokens: str) -> str:
        missing = self.tokens - set(tokens.keys())
        if len(missing):
            raise CodeTemplateError(
                f"Cannot instantiate template: {self.name} - missing required tokens: {missing}"
            )

        result = copy.copy(self.template_string)
        for token, value in tokens.items():
            result = re.sub(r"\%{}\%".format(token), value, result)

        return result



"""Wrapper objects for Unreal Engine builds."""

# Future Standard Library
from __future__ import annotations

# Standard Library
import platform
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, Iterable, Optional

if TYPE_CHECKING:
    # CrazyHusk
    from crazyhusk.engine import UnrealEngine


class Buildable(ABC):
    @abstractmethod
    def is_buildable(self) -> bool:
        ...

    @abstractmethod
    def get_build_command(
        self,
        target: Optional[str] = None,
        configuration: Optional[str] = None,
        platform: Optional[str] = None,
        *extra_switches: str,
        **extra_parameters: str,
    ) -> Iterable[str]:
        ...

    @property
    @abstractmethod
    def engine(self) -> Optional[UnrealEngine]:
        ...

    def is_valid_build_target(self, target: str) -> bool:
        return target in {"Game", "Editor", "Server"}

    def is_valid_build_platform(self, platform: str) -> bool:
        return platform in {
            "Win32",
            "Win64",
            "Linux",
            "LinuxAArch64",
            "Mac",
            "Android",
            "IOS",
        }

    def is_valid_build_configuration(self, configuration: str) -> bool:
        return configuration in {
            "Debug",
            "DebugGame",
            "Development",
            "Shipping",
            "Test",
        }

    def default_local_platform(self) -> str:
        local_system = platform.system()
        if local_system == "Windows":
            return "Win64"
        elif local_system == "Linux":
            return "Linux"
        elif local_system == "Darwin":
            return "Mac"
        raise NotImplementedError(
            f"Default build platform for {local_system} not defined."
        )

    def default_build_target(self) -> str:
        return "Editor"

    def default_build_configuration(self) -> str:
        return "Development"


class UnrealBuild(object):
    buildable: Buildable
    __target: str
    __configuration: str
    __platform: str

    def __init__(
        self,
        buildable: Buildable,
        target: Optional[str] = None,
        configuration: Optional[str] = None,
        build_platform: Optional[str] = None,
    ) -> None:
        self.buildable = buildable
        if target is None:
            self.target = self.buildable.default_build_target()
        else:
            self.target = target

        if configuration is None:
            self.configuration = self.buildable.default_build_configuration()
        else:
            self.configuration = configuration

        if build_platform is None:
            self.platform = self.buildable.default_local_platform()
        else:
            self.platform = build_platform

    @property
    def target(self) -> str:
        return self.__target

    @target.setter
    def target(self, value: str) -> None:
        if self.buildable.is_valid_build_target(value):
            self.__target = value

    @property
    def configuration(self) -> str:
        return self.__configuration

    @configuration.setter
    def configuration(self, value: str) -> None:
        if self.buildable.is_valid_build_configuration(value):
            self.__configuration = value

    @property
    def platform(self) -> str:
        return self.__platform

    @platform.setter
    def platform(self, value: str) -> None:
        if self.buildable.is_valid_build_platform(value):
            self.__platform = value

    def run(
        self,
        *extra_switches: str,
        **extra_parameters: str,
    ) -> int:
        if not self.buildable.is_buildable():
            raise ValueError(f"Buildable: {self.buildable!r} cannot be built.")

        if self.buildable.engine is None:
            raise ValueError(
                f"Buildable: {self.buildable!r} could not resolve a valid UnrealEngine."
            )

        with self.buildable.engine:
            return self.buildable.engine.run(
                *self.buildable.get_build_command(
                    self.target,
                    self.configuration,
                    self.platform,
                    *extra_switches,
                    **extra_parameters,
                ),
                expected_retcodes={0, 2},
            )



# Standard Library
import glob
import sqlite3
from contextlib import closing


def main():
    for coverage_db in glob.iglob("./.coverage.*"):
        with closing(sqlite3.connect(coverage_db)) as connection:
            with closing(connection.cursor()) as cursor:
                cursor.execute("UPDATE file SET path=REPLACE(path,'\\','/')")
                connection.commit()


if __name__ == "__main__":
    main()



"""Sphinx configuration."""
project = "CrazyHusk"
author = "Nick Haines"
copyright = "2022, Nick Haines"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_click",
    "myst_parser",
]
autodoc_typehints = "description"
html_theme = "furo"



"""Monkeypatching for module-level issues, such as mocking winreg"""

# Future Standard Library
from __future__ import annotations

# Standard Library
import json
import os
import sys
from types import ModuleType, TracebackType
from typing import Any, Callable, Dict, Iterable, Optional, Type
from xml.etree.ElementTree import Element

# Third Party
import pytest

# CrazyHusk
from crazyhusk import code, config, engine, logs, module, plugin, project


@pytest.fixture(scope="function")
def mock_winreg(monkeypatch: Any) -> None:
    class HKEYType:
        def __bool__(self) -> bool:
            ...

        def __int__(self) -> int:
            ...

        def __enter__(self) -> HKEYType:
            ...

        def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[Type[BaseException]],
            exc_tb: Optional[Type[TracebackType]],
        ) -> bool | None:
            ...

        def Close(self) -> None:
            ...

        def Detach(self) -> int:
            ...

    def OpenKey(key: Any, sub_key: str, reserved: int = 0, access: int = 0) -> HKEYType:
        return HKEYType()

    def OpenKeyEx(
        key: Any, sub_key: str, reserved: int = 0, access: int = 0
    ) -> HKEYType:
        return HKEYType()

    def EnumValue(key: HKEYType, index: int) -> tuple[str, Any, int]:
        raise OSError()

    def QueryValueEx(__key: HKEYType, __name: str) -> tuple[Any, int]:
        raise OSError()

    module = ModuleType("winreg")
    setattr(module, "OpenKey", OpenKey)
    setattr(module, "OpenKeyEx", OpenKeyEx)
    setattr(module, "EnumValue", EnumValue)
    setattr(module, "QueryValueEx", QueryValueEx)
    setattr(module, "HKEY_CURRENT_USER", 1)
    setattr(module, "HKEY_LOCAL_MACHINE", 1)
    monkeypatch.setitem(sys.modules, "winreg", module)


@pytest.fixture(scope="function")
def null_code_template() -> code.CodeTemplate:
    yield code.CodeTemplate(None)


@pytest.fixture(scope="function")
def empty_code_template() -> code.CodeTemplate:
    yield code.CodeTemplate("")


@pytest.fixture(scope="function")
def basic_code_template() -> code.CodeTemplate:
    yield code.CodeTemplate("Basic", "%TEST_TOKEN%")


@pytest.fixture(scope="function")
def multiline_basic_code_template() -> code.CodeTemplate:
    yield code.CodeTemplate("MultilineBasic", "//%TEST_TOKEN%\n//%TEST_TOKEN%")


@pytest.fixture(scope="function")
def multiline_multitoken_code_template() -> code.CodeTemplate:
    yield code.CodeTemplate(
        "MultilineMultitoken", r"//%TEST_TOKEN_1%\n//%TEST_TOKEN_2%"
    )


@pytest.fixture(scope="function")
def empty_parser() -> config.UnrealConfigParser:
    yield config.UnrealConfigParser()


@pytest.fixture(scope="function")
def version_empty() -> engine.UnrealVersion:
    yield engine.UnrealVersion()


@pytest.fixture(scope="function")
def version_5_0() -> engine.UnrealVersion:
    version = engine.UnrealVersion()
    version.major = 5
    yield version


@pytest.fixture(scope="function")
def version_26_0() -> engine.UnrealVersion:
    version = engine.UnrealVersion()
    version.minor = 26
    return version


@pytest.fixture(scope="function")
def version_26_1() -> engine.UnrealVersion:
    version = engine.UnrealVersion()
    version.minor = 26
    version.patch = 1
    return version


@pytest.fixture(scope="function")
def version_26_1_123456() -> engine.UnrealVersion:
    version = engine.UnrealVersion()
    version.minor = 26
    version.patch = 1
    version.changelist = 123456
    return version


@pytest.fixture(scope="function")
def version_26_1_234567() -> engine.UnrealVersion:
    version = engine.UnrealVersion()
    version.minor = 26
    version.patch = 1
    version.changelist = 234567
    return version


@pytest.fixture(scope="function")
def version_26_1_123456_branch() -> engine.UnrealVersion:
    version = engine.UnrealVersion()
    version.minor = 26
    version.patch = 1
    version.changelist = 123456
    version.branch = "++UE4+Release-4.26"
    return version


@pytest.fixture(scope="function")
def version_egl_4_26_2() -> engine.UnrealVersion:
    version = engine.UnrealVersion()
    version.minor = 26
    version.patch = 2
    version.changelist = 15973114
    version.branch = "++UE4+Release-4.26"
    return version


@pytest.fixture(scope="function")
def engine_empty(tmp_path: Any) -> engine.UnrealEngine:
    yield engine.UnrealEngine(tmp_path / "Empty")


@pytest.fixture(scope="function")
def engine_local() -> engine.UnrealEngine:
    yield engine.UnrealEngine(".")


@pytest.fixture(scope="function")
def engine_empty_version_empty(
    tmp_path: Any, version_empty: engine.UnrealVersion
) -> engine.UnrealEngine:
    tmp_engine_dir = tmp_path / "EmptyVersion"
    tmp_engine_dir.mkdir()
    engine_dir = tmp_engine_dir / "Engine"
    engine_dir.mkdir()
    build_dir = engine_dir / "Build"
    build_dir.mkdir()
    version_file = build_dir / "Build.version"
    version_file.write_text(json.dumps(version_empty.to_dict()))
    source_file = build_dir / "SourceDistribution.txt"
    source_file.write_text("")
    yield engine.UnrealEngine(tmp_engine_dir)


@pytest.fixture(scope="function")
def engine_empty_version_egl_4_26_2(
    tmp_path: Any, version_egl_4_26_2: engine.UnrealVersion
) -> engine.UnrealEngine:
    tmp_engine_dir = tmp_path / "EGL4.26.2"
    tmp_engine_dir.mkdir()
    engine_dir = tmp_engine_dir / "Engine"
    engine_dir.mkdir()
    build_dir = engine_dir / "Build"
    build_dir.mkdir()
    version_file = build_dir / "Build.version"
    version_file.write_text(json.dumps(version_egl_4_26_2.to_dict()))
    installed_file = build_dir / "InstalledBuild.txt"
    installed_file.write_text("")
    config_dir = engine_dir / "Config"
    config_dir.mkdir()
    base_config_file = config_dir / "Base.ini"
    base_config_file.write_text("")
    yield engine.UnrealEngine(tmp_engine_dir)


@pytest.fixture(scope="function")
def null_filter_engine_run() -> logs.FilterEngineRun:
    yield logs.FilterEngineRun(None)


@pytest.fixture(scope="function")
def empty_filter_engine_run() -> logs.FilterEngineRun:
    yield logs.FilterEngineRun("")


@pytest.fixture(scope="function")
def one_arg_filter_engine_run() -> logs.FilterEngineRun:
    yield logs.FilterEngineRun("", "arg")


@pytest.fixture(scope="function")
def multi_arg_filter_engine_run() -> logs.FilterEngineRun:
    yield logs.FilterEngineRun("", "arg1", "arg2", "arg3")


@pytest.fixture(scope="function")
def null_module_descriptor() -> module.ModuleDescriptor:
    yield module.ModuleDescriptor()


@pytest.fixture(scope="function")
def empty_module_descriptor() -> module.ModuleDescriptor:
    descriptor = module.ModuleDescriptor()
    descriptor.name = ""
    yield descriptor


@pytest.fixture(scope="function")
def invalid_hosttype_module_descriptor() -> module.ModuleDescriptor:
    descriptor = module.ModuleDescriptor()
    descriptor.name = "InvalidHostType"
    descriptor.host_type = "Invalid"
    yield descriptor


@pytest.fixture(scope="function")
def invalid_loadingphase_module_descriptor() -> module.ModuleDescriptor:
    descriptor = module.ModuleDescriptor()
    descriptor.name = "InvalidLoadingPhase"
    descriptor.host_type = "Runtime"
    descriptor.loading_phase = "Invalid"
    yield descriptor


@pytest.fixture(scope="function")
def default_valid_module_descriptor() -> module.ModuleDescriptor:
    descriptor = module.ModuleDescriptor()
    descriptor.name = "DefaultValid"
    descriptor.host_type = "Runtime"
    descriptor.loading_phase = "Default"
    yield descriptor


@pytest.fixture(scope="function")
def null_name_plugin_descriptor() -> plugin.PluginDescriptor:
    _plugin = plugin.PluginDescriptor()
    _plugin.friendly_name = None
    yield _plugin


@pytest.fixture(scope="function")
def empty_name_plugin_descriptor() -> plugin.PluginDescriptor:
    yield plugin.PluginDescriptor()


@pytest.fixture(scope="function")
def null_version_plugin_descriptor() -> plugin.PluginDescriptor:
    _plugin = plugin.PluginDescriptor()
    _plugin.friendly_name = "NullVersion"
    _plugin.version_name = None
    yield _plugin


@pytest.fixture(scope="function")
def empty_version_plugin_descriptor() -> plugin.PluginDescriptor:
    _plugin = plugin.PluginDescriptor()
    _plugin.friendly_name = "EmptyVersion"
    yield _plugin


@pytest.fixture(scope="function")
def basic_plugin_descriptor() -> plugin.PluginDescriptor:
    _plugin = plugin.PluginDescriptor()
    _plugin.friendly_name = "Basic"
    _plugin.version_name = "1.0"
    yield _plugin


@pytest.fixture(scope="function")
def basic_plugin_descriptor_withmodule(
    basic_plugin_descriptor: plugin.PluginDescriptor,
    default_valid_module_descriptor: module.ModuleDescriptor,
) -> plugin.PluginDescriptor:
    basic_plugin_descriptor.add_module(default_valid_module_descriptor)
    yield basic_plugin_descriptor


@pytest.fixture(scope="function")
def basic_plugin_descriptor_withmodule_dict(
    basic_plugin_descriptor: plugin.PluginDescriptor,
    default_valid_module_descriptor: module.ModuleDescriptor,
) -> Iterable[Dict[str, Any]]:
    dct = basic_plugin_descriptor.to_dict()
    dct["Modules"].append(default_valid_module_descriptor.to_dict())
    yield dct


@pytest.fixture(scope="function")
def null_plugin_reference_descriptor() -> plugin.PluginReferenceDescriptor:
    yield plugin.PluginReferenceDescriptor()


@pytest.fixture(scope="function")
def empty_plugin_reference_descriptor() -> plugin.PluginReferenceDescriptor:
    ref = plugin.PluginReferenceDescriptor()
    ref.name = ""
    yield ref


@pytest.fixture(scope="function")
def basic_plugin_reference_descriptor() -> plugin.PluginReferenceDescriptor:
    ref = plugin.PluginReferenceDescriptor()
    ref.name = "Basic"
    yield ref


@pytest.fixture(scope="function")
def empty_unreal_plugin() -> plugin.UnrealPlugin:
    yield plugin.UnrealPlugin("")


@pytest.fixture(scope="function")
def local_dir_unreal_plugin() -> plugin.UnrealPlugin:
    yield plugin.UnrealPlugin(".")


@pytest.fixture(scope="function")
def invalid_file_unreal_plugin() -> plugin.UnrealPlugin:
    yield plugin.UnrealPlugin("./test.txt")


@pytest.fixture(scope="function")
def invalid_file_unreal_plugin_realpath() -> plugin.UnrealPlugin:
    yield plugin.UnrealPlugin(os.path.realpath("./test.txt"))


@pytest.fixture(scope="function")
def empty_file_content_unreal_plugin(tmp_path: Any) -> plugin.UnrealPlugin:
    plugin_file = tmp_path / "Invalid.uplugin"
    plugin_file.write_text("")
    yield plugin.UnrealPlugin(plugin_file)


@pytest.fixture(scope="function")
def basic_unreal_plugin(
    tmp_path: Any, basic_plugin_descriptor_withmodule_dict: Iterable[Dict[str, Any]]
) -> plugin.UnrealPlugin:
    plugin_file = tmp_path / "Basic.uplugin"
    plugin_file.write_text(json.dumps(basic_plugin_descriptor_withmodule_dict))
    yield plugin.UnrealPlugin(plugin_file)


@pytest.fixture(scope="function")
def null_association_project_descriptor() -> project.ProjectDescriptor:
    yield project.ProjectDescriptor()


@pytest.fixture(scope="function")
def empty_association_project_descriptor() -> project.ProjectDescriptor:
    _project = project.ProjectDescriptor()
    _project.engine_association = ""
    yield _project


@pytest.fixture(scope="function")
def basic_project_descriptor() -> project.ProjectDescriptor:
    _project = project.ProjectDescriptor()
    _project.description = "Basic"
    _project.engine_association = ""
    yield _project


@pytest.fixture(scope="function")
def basic_project_descriptor_withmodule(
    basic_project_descriptor: project.ProjectDescriptor,
    default_valid_module_descriptor: module.ModuleDescriptor,
) -> project.ProjectDescriptor:
    basic_project_descriptor.add_module(default_valid_module_descriptor)
    yield basic_project_descriptor


@pytest.fixture(scope="function")
def basic_project_descriptor_withmodule_dict(
    basic_project_descriptor: project.ProjectDescriptor,
    default_valid_module_descriptor: module.ModuleDescriptor,
) -> Iterable[Dict[str, Any]]:
    dct = basic_project_descriptor.to_dict()
    dct["Modules"].append(default_valid_module_descriptor.to_dict())
    yield dct


@pytest.fixture(scope="function")
def null_filename_unreal_project() -> project.UnrealProject:
    yield project.UnrealProject(None)


@pytest.fixture(scope="function")
def empty_filename_unreal_project() -> project.UnrealProject:
    yield project.UnrealProject("")


@pytest.fixture(scope="function")
def invalid_filename_unreal_project() -> project.UnrealProject:
    yield project.UnrealProject("MyProject.txt")


@pytest.fixture(scope="function")
def empty_file_unreal_project() -> project.UnrealProject:
    yield project.UnrealProject("MyProject.uproject")


@pytest.fixture(scope="function")
def empty_file_content_unreal_project(tmp_path: Any) -> project.UnrealProject:
    project_file = tmp_path / "MyProject.uproject"
    project_file.write_text("")
    yield project.UnrealProject(project_file)


@pytest.fixture(scope="function")
def null_engine_unreal_project(tmp_path: Any) -> project.UnrealProject:
    project.entry_points = lambda: {}
    project_file = tmp_path / "MyProject.uproject"
    project_file.write_text('{"EngineAssociation":"123456"}')
    _project = project.UnrealProject(project_file)
    _project.engine = None
    yield _project


@pytest.fixture(scope="function")
def basic_unreal_project(
    tmp_path: Any, basic_project_descriptor_withmodule_dict: Iterable[Dict[str, Any]]
) -> project.UnrealProject:
    project_file = tmp_path / "MyProject.uproject"
    project_file.write_text(json.dumps(basic_project_descriptor_withmodule_dict))
    yield project.UnrealProject(project_file)


@pytest.fixture(scope="function")
def basic_unreal_project_realpath(
    basic_project_descriptor_withmodule_dict: Iterable[Dict[str, Any]]
) -> project.UnrealProject:
    project_file = os.path.realpath("./MyProject.uproject")
    engine_dir = os.path.realpath("../Engine")
    os.makedirs(engine_dir, exist_ok=True)
    with open(project_file, "w", encoding="utf-8") as _file:
        json.dump(basic_project_descriptor_withmodule_dict, _file)
    yield project.UnrealProject(project_file)
    os.remove(project_file)
    os.removedirs(engine_dir)


@pytest.fixture(scope="function")
def null_report_file() -> None:
    return None


@pytest.fixture(scope="function")
def empty_filename_report_file() -> str:
    return ""


@pytest.fixture(scope="function")
def non_json_report_file(tmp_path: Any) -> str:
    non_json_file = tmp_path / "non-json.file"
    non_json_file.write_text("")
    return str(non_json_file)


@pytest.fixture(scope="function")
def empty_json_report_file(tmp_path: Any) -> str:
    empty_json_file = tmp_path / "empty.json"
    empty_json_file.write_text("")
    return str(empty_json_file)


@pytest.fixture(scope="function")
def list_json_report_file(tmp_path: Any) -> str:
    list_json_file = tmp_path / "list.json"
    list_json_file.write_text("[]")
    return str(list_json_file)


@pytest.fixture(scope="function")
def empty_dict_json_report_file(tmp_path: Any) -> str:
    empty_dict_json_file = tmp_path / "empty_dict.json"
    empty_dict_json_file.write_text("{}")
    return str(empty_dict_json_file)


@pytest.fixture(scope="function")
def xml_filename_report_file(tmp_path: Any) -> str:
    return str(tmp_path / "empty.xml")


@pytest.fixture(scope="function")
def xml_filename_mkdirs_report_file(tmp_path: Any) -> str:
    return str(tmp_path / "test/empty.xml")


@pytest.fixture(scope="function")
def null_test_suites() -> None:
    return None


@pytest.fixture(scope="function")
def emptystring_test_suites() -> str:
    return ""


@pytest.fixture(scope="function")
def empty_element_test_suites() -> Element:
    return Element("")


@pytest.fixture(scope="function")
def basic_element_test_suites() -> Element:
    return Element("testsuites")


@pytest.fixture(scope="function")
def null_entry_point() -> object:
    class NullEntryPoint:
        def __init__(self) -> None:
            self.name = "test"

        def load(*args: Any) -> None:
            return None

    yield NullEntryPoint()


@pytest.fixture(scope="function")
def test_entry_point() -> object:
    def test(*args: Any) -> None:
        return

    class TestEntryPoint:
        def __init__(self) -> None:
            self.name = "test"

        def load(*args: Any) -> Callable[[], None]:
            return test

    yield TestEntryPoint()


@pytest.fixture(scope="function")
def basic_datfile(tmp_path: Any) -> Iterable[str]:
    yield json.dumps(
        {
            "InstallationList": [
                {
                    "InstallLocation": "C:\\EpicStore\\UE_4.26",
                    "NamespaceId": "ue",
                    "ItemId": "3ddb1bad6e004b99a7192c1a29f2318a",
                    "ArtifactId": "UE_4.26",
                    "AppVersion": "4.26.2-15973114+++UE4+Release-4.26-Windows",
                    "AppName": "UE_4.26",
                },
            ]
        }
    )



# Standard Library
from typing import Any, Type

# Third Party
import pytest

# CrazyHusk
from crazyhusk import module


def test_module_descriptor_init(
    null_module_descriptor: module.ModuleDescriptor,
) -> None:
    assert null_module_descriptor.name is None
    assert null_module_descriptor.host_type is None
    assert null_module_descriptor.loading_phase is None
    assert null_module_descriptor.additional_dependencies == []
    assert null_module_descriptor.blacklist_platforms == []
    assert null_module_descriptor.blacklist_programs == []
    assert null_module_descriptor.blacklist_target_configurations == []
    assert null_module_descriptor.blacklist_targets == []
    assert null_module_descriptor.whitelist_platforms == []
    assert null_module_descriptor.whitelist_programs == []
    assert null_module_descriptor.whitelist_target_configurations == []
    assert null_module_descriptor.whitelist_targets == []


def test_module_descriptor_repr(
    null_module_descriptor: module.ModuleDescriptor,
) -> None:
    assert repr(null_module_descriptor) == "<ModuleDescriptor None>"


def test_module_descriptor_hash(
    null_module_descriptor: module.ModuleDescriptor,
) -> None:
    assert hash(null_module_descriptor) == hash(None)


def test_module_descriptor_equality(
    null_module_descriptor: module.ModuleDescriptor,
    empty_module_descriptor: module.ModuleDescriptor,
) -> None:
    assert null_module_descriptor == null_module_descriptor
    assert null_module_descriptor != empty_module_descriptor


@pytest.mark.parametrize(
    "module_descriptor_fixture,expected",
    [
        ("null_module_descriptor", False),
        ("empty_module_descriptor", False),
        ("invalid_hosttype_module_descriptor", False),
        ("invalid_loadingphase_module_descriptor", False),
        ("default_valid_module_descriptor", True),
    ],
)
def test_module_descriptor_is_valid(
    module_descriptor_fixture: str, expected: bool, request: Any
) -> None:
    module_descriptor = request.getfixturevalue(module_descriptor_fixture)
    assert module_descriptor.is_valid() is expected


@pytest.mark.parametrize(
    "module_descriptor_fixture,expected_type",
    [
        ("null_module_descriptor", dict),
        ("empty_module_descriptor", dict),
        ("invalid_hosttype_module_descriptor", dict),
        ("invalid_loadingphase_module_descriptor", dict),
        ("default_valid_module_descriptor", module.ModuleDescriptor),
    ],
)
def test_module_descriptor_to_object(
    module_descriptor_fixture: str, expected_type: Type[Any], request: Any
) -> None:
    module_descriptor = request.getfixturevalue(module_descriptor_fixture)
    dct = module_descriptor.to_dict()
    assert isinstance(module.ModuleDescriptor.to_object(dct), expected_type)



# Standard Library
import configparser

# Third Party
import pytest

# CrazyHusk
from crazyhusk import config


def test_config_init(empty_parser: config.UnrealConfigParser) -> None:
    assert isinstance(empty_parser, configparser.RawConfigParser)


@pytest.mark.parametrize(
    "input_string,output_string",
    [
        ("AxisConfig", "AxisConfig"),
        ("+AxisConfig", "AxisConfig"),
        ("-AxisConfig", "AxisConfig"),
        (".AxisConfig", "AxisConfig"),
        ("!AxisConfig", "AxisConfig"),
    ],
)
def test_config_optionxform(
    empty_parser: config.UnrealConfigParser, input_string: str, output_string: str
) -> None:
    assert empty_parser.optionxform(input_string) == output_string



# Standard Library
from typing import Any

# Third Party
import pytest

# CrazyHusk
import crazyhusk.windows
from crazyhusk.engine import UnrealEngine
from crazyhusk.windows import engine


def test_find_egl_engine_windows_no_datfile(tmp_path: Any, monkeypatch: Any) -> None:
    dat_file = tmp_path / "LauncherInstalled.dat"
    engine.DAT_FILE = dat_file
    monkeypatch.setattr("platform.system", lambda: "NotWindows")
    assert engine.find_egl_engine_windows("") is None
    monkeypatch.setattr("platform.system", lambda: "Windows")
    assert engine.find_egl_engine_windows("") is None


def test_find_egl_engine_windows_with_datfile(
    basic_datfile: str, tmp_path: Any, monkeypatch: Any
) -> None:
    dat_file = tmp_path / "LauncherInstalled.dat"
    dat_file.write_text(basic_datfile)
    engine.DAT_FILE = dat_file

    monkeypatch.setattr("platform.system", lambda: "NotWindows")
    assert engine.find_egl_engine_windows("4.26") is None

    monkeypatch.setattr("platform.system", lambda: "Windows")
    assert isinstance(engine.find_egl_engine_windows("4.26"), UnrealEngine)
    assert engine.find_egl_engine_windows("4.27") is None


def test_list_egl_engine_windows_no_datfile(tmp_path: Any, monkeypatch: Any) -> None:
    dat_file = tmp_path / "LauncherInstalled.dat"
    engine.DAT_FILE = dat_file
    monkeypatch.setattr("platform.system", lambda: "NotWindows")
    for _engine in engine.list_egl_engines_windows():
        assert _engine is None
    monkeypatch.setattr("platform.system", lambda: "Windows")
    for _engine in engine.list_egl_engines_windows():
        assert _engine is None


def test_list_egl_engine_windows_with_datfile(
    basic_datfile: str, tmp_path: Any, monkeypatch: Any
) -> None:
    dat_file = tmp_path / "LauncherInstalled.dat"
    dat_file.write_text(basic_datfile)
    engine.DAT_FILE = dat_file
    monkeypatch.setattr("platform.system", lambda: "NotWindows")
    for _engine in engine.list_egl_engines_windows():
        assert _engine is None
    monkeypatch.setattr("platform.system", lambda: "Windows")
    for _engine in engine.list_egl_engines_windows():
        assert isinstance(_engine, UnrealEngine)


def test_registry_nonwindows(monkeypatch: Any, mock_winreg: Any) -> None:
    monkeypatch.setattr("platform.system", lambda: "NotWindows")
    assert engine.find_registered_engines_windows("random_string") is None
    assert len(list(engine.list_registered_engines_windows())) == 0


def test_registry_windows(monkeypatch: Any, mock_winreg: Any) -> None:
    monkeypatch.setattr("platform.system", lambda: "Windows")
    assert engine.find_registered_engines_windows("random_string") is None
    assert len(list(engine.list_registered_engines_windows())) == 0


def test_resolve_executable_path_windows(
    monkeypatch: Any, engine_local: engine.UnrealEngine
) -> None:
    monkeypatch.setattr("platform.system", lambda: "NotWindows")
    assert engine.resolve_executable_path_windows(None, None) is None
    monkeypatch.setattr("platform.system", lambda: "Windows")
    with pytest.raises(TypeError):
        assert engine.resolve_executable_path_windows(None, None) is None

    with pytest.raises(AttributeError):
        assert engine.resolve_executable_path_windows(engine_local, None) is None

    assert engine.resolve_executable_path_windows(engine_local, "") is None
    assert engine.resolve_executable_path_windows(engine_local, "random_string") is None
    assert (
        engine.resolve_executable_path_windows(engine_local, "AutomationTool")
        is not None
    )
    assert (
        engine.resolve_executable_path_windows(engine_local, "automationtool")
        is not None
    )
    assert (
        engine.resolve_executable_path_windows(engine_local, "swarmagent") is not None
    )
    assert (
        engine.resolve_executable_path_windows(engine_local, "swarmcoordinator")
        is not None
    )
    assert (
        engine.resolve_executable_path_windows(engine_local, "unrealbuildtool")
        is not None
    )
    assert engine.resolve_executable_path_windows(engine_local, "ue4editor") is not None
    assert (
        engine.resolve_executable_path_windows(engine_local, "ue4editor-cmd")
        is not None
    )
    assert engine.resolve_executable_path_windows(engine_local, "unrealpak") is not None



# Standard Library
import logging
from typing import Any, Optional, Tuple

# Third Party
import pytest

# CrazyHusk
from crazyhusk import logs


@pytest.mark.parametrize(
    "filter_engine_run_fixture,executable,cmd_args",
    [
        ("null_filter_engine_run", None, ()),
        ("empty_filter_engine_run", "", ()),
        ("one_arg_filter_engine_run", "", ("arg",)),
        (
            "multi_arg_filter_engine_run",
            "",
            (
                "arg1",
                "arg2",
                "arg3",
            ),
        ),
    ],
)
def test_filter_engine_run_init(
    filter_engine_run_fixture: str,
    executable: Optional[str],
    cmd_args: Tuple[str],
    request: Any,
) -> None:
    filter_engine_run = request.getfixturevalue(filter_engine_run_fixture)
    assert filter_engine_run.executable == executable
    assert isinstance(filter_engine_run.cmd_args, tuple)
    assert filter_engine_run.cmd_args == cmd_args


@pytest.mark.parametrize(
    "filter_engine_run_fixture,executable,cmd_args",
    [
        ("null_filter_engine_run", None, ()),
        ("empty_filter_engine_run", "", ()),
        ("one_arg_filter_engine_run", "", ("arg",)),
        (
            "multi_arg_filter_engine_run",
            "",
            (
                "arg1",
                "arg2",
                "arg3",
            ),
        ),
    ],
)
def test_filter_engine_run_filter(
    filter_engine_run_fixture: str,
    executable: Optional[str],
    cmd_args: Tuple[str],
    request: Any,
    caplog: Any,
) -> None:
    filter_engine_run = request.getfixturevalue(filter_engine_run_fixture)
    logger = logging.getLogger("test_filter_engine_run_filter")
    logger.addFilter(filter_engine_run)
    with caplog.at_level(logging.INFO):
        logger.info("test")
        for record in caplog.records:
            assert record.executable == executable
            assert record.cmd_args == cmd_args


@pytest.mark.parametrize(
    "log_string,levelno,levelname,filename,linenumber,colnumber,sub_msg",
    [
        (
            "@progress 'Compiling C++ source code...' 59%",
            logging.INFO,
            "INFO",
            "test_logs.py",
            None,
            None,
            None,
        ),
        (
            "[8/17] MainMenuPlayerController.gen.cpp",
            logging.INFO,
            "INFO",
            "test_logs.py",
            None,
            None,
            None,
        ),
        (
            "/path/to/myfile.cpp(32): error C3861: 'assert': identifier not found",
            logging.ERROR,
            "ERROR",
            "/path/to/myfile.cpp",
            "32",
            None,
            "'assert': identifier not found",
        )
        # TODO: find  appropriate warning message examples, including ones with column numbers
    ],
)
def test_filter_ubt_warnings(
    log_string: str,
    levelno: int,
    levelname: str,
    filename: Optional[str],
    linenumber: Optional[str],
    colnumber: Optional[str],
    sub_msg: Optional[str],
    caplog: Any,
) -> None:
    logger = logging.getLogger("test_filter_ubt_warnings")
    logger.addFilter(logs.FilterUBTWarnings())
    with caplog.at_level(logging.INFO):
        logger.info(log_string)
        for record in caplog.records:
            assert record.levelno == levelno
            assert record.levelname == levelname
            assert record.filename == filename
            if linenumber is not None:
                assert record.linenumber == linenumber
            if colnumber is not None:
                assert record.colnumber == colnumber
            if sub_msg is not None:
                assert record.sub_msg == sub_msg


@pytest.mark.parametrize(
    "log_string,levelno,levelname,created,module,sub_msg",
    [
        (
            "LogPluginManager: Mounting plugin SunPosition",
            logging.INFO,
            "INFO",
            None,
            None,
            None,
        ),
        (
            "[2021.11.29-21.06.49:745][  0]LogConfig: Setting CVar [[r.BloomQuality:5]]",
            logging.INFO,
            "INFO",
            1638220009,
            "LogConfig",
            "Setting CVar [[r.BloomQuality:5]]",
        ),
    ],
)
def test_filter_ue4_logs(
    log_string: str,
    levelno: int,
    levelname: str,
    created: Optional[int],
    module: Optional[str],
    sub_msg: Optional[str],
    caplog: Any,
) -> None:
    logger = logging.getLogger("test_filter_ue4_logs")
    logger.addFilter(logs.FilterUE4Logs())
    with caplog.at_level(logging.INFO):
        logger.info(log_string)
        for record in caplog.records:
            assert record.levelno == levelno
            assert record.levelname == levelname
            if created is not None:
                assert record.created == created
            if module is not None:
                assert record.module == module
            if sub_msg is not None:
                assert record.sub_msg == sub_msg



"""Object wrappers for working with Unreal Engine config files."""

# Standard Library
import configparser
import re

CONFIG_CATEGORIES = frozenset(
    [
        "Compat",
        "DeviceProfiles",
        "Editor",
        "EditorGameAgnostic",
        "EditorKeyBindings",
        "EditorLayout",
        "EditorPerProjectUserSettings",
        "EditorSettings",
        "Engine",
        "Game",
        "GameUserSettings",
        "Hardware",
        "Input",
        "InstallBundle",
        "Lightmass",
        "PakFileRules",
        "RuntimeOptions",
        "Scalability",
        "SourceControlSettings",
    ]
)


class UnrealConfigError(Exception):
    """Custom exception representing errors encountered with Unreal config files."""


class UnrealConfigParser(configparser.RawConfigParser):
    """Object wrapper representing a configuration stack."""

    RE_OPTION_SPECIALCHARS = re.compile(r"^([+-.!])")

    def __init__(self) -> None:
        super().__init__(strict=False)

    def optionxform(self, optionstr: str) -> str:
        return UnrealConfigParser.RE_OPTION_SPECIALCHARS.sub("", optionstr)



"""Wrapper objects for Unreal plugins."""

# Future Standard Library
from __future__ import annotations

# Standard Library
import glob
import json
import os
from typing import Any, Dict, Iterable, List, Optional, Union

# CrazyHusk
from crazyhusk.build import Buildable
from crazyhusk.engine import UnrealEngine

try:
    # Standard Library
    from importlib.metadata import entry_points  # type:ignore
except ImportError:
    # Third Party
    from importlib_metadata import entry_points  # type:ignore

# CrazyHusk
from crazyhusk.code import CodeTemplate
from crazyhusk.module import ModuleDescriptor

__all__ = ["UnrealPlugin"]


class UnrealPluginError(Exception):
    """Custom exception representing errors encountered with Unreal Plugins."""


class PluginDescriptor(object):
    """Object wrapper representation of a uplugin file, equivalent to serialization method used with FPluginDescriptor.

    https://docs.unrealengine.com/en-US/API/Runtime/Projects/FPluginDescriptor/index.html
    """

    def __init__(self) -> None:
        """Initialize a PluginDescriptor."""
        self.can_contain_content: bool = False
        self.explicitly_loaded: bool = False
        self.installed: bool = False
        self.is_beta_version: bool = False
        self.is_experimental_version: bool = False
        self.is_hidden: bool = False
        self.is_plugin_extension: bool = False
        self.requires_build_platform: bool = False
        self.category: str = ""
        self.created_by: str = ""
        self.created_by_url: str = ""
        self.description: str = ""
        self.docs_url: str = ""
        self.editor_custom_virtual_path: str = ""
        self.enabled_by_default: bool = False
        self.engine_version: str = ""
        self.friendly_name: str = ""
        self.localization_targets: List[Any] = []
        self.marketplace_url: str = ""
        self.__modules: List[Any] = []
        self.parent_plugin_name: str = ""
        self.__plugins: List[Any] = []
        self.post_build_steps: Optional[Any] = None
        self.pre_build_steps: Optional[Any] = None
        self.supported_programs: List[Any] = []
        self.supported_target_platforms: List[Any] = []
        self.support_url: str = ""
        self.version: int = 1
        self.version_name: str = ""

    def __repr__(self) -> str:
        """Python interpreter representation of PluginDescriptor."""
        return f"<PluginDescriptor {self.friendly_name} version {self.version_name}>"

    @property
    def modules(self) -> Iterable[Union[ModuleDescriptor, Dict[str, Any]]]:
        for module in self.__modules:
            if isinstance(module, dict):
                yield ModuleDescriptor.to_object(module)
            else:
                yield module

    @property
    def plugins(self) -> Iterable[Union[PluginReferenceDescriptor, Dict[str, Any]]]:
        for plugin in self.__plugins:
            yield PluginReferenceDescriptor.to_object(plugin)

    @staticmethod
    def to_object(dct: Dict[str, Any]) -> Union[PluginDescriptor, Dict[str, Any]]:
        descriptor = PluginDescriptor()
        descriptor.can_contain_content = dct.get("CanContainContent", False)
        descriptor.explicitly_loaded = dct.get("ExplicitlyLoaded", False)
        descriptor.installed = dct.get("Installed", False)
        descriptor.is_beta_version = dct.get("IsBetaVersion", False)
        descriptor.is_experimental_version = dct.get("IsExperimentalVersion", False)
        descriptor.is_hidden = dct.get("IsHidden", False)
        descriptor.is_plugin_extension = dct.get("IsPluginExtension", False)
        descriptor.requires_build_platform = dct.get("RequiresBuildPlatform", False)
        descriptor.category = dct.get("Category", "")
        descriptor.created_by = dct.get("CreatedBy", "")
        descriptor.created_by_url = dct.get("CreatedByURL", "")
        descriptor.description = dct.get("Description", "")
        descriptor.docs_url = dct.get("DocsURL", "")
        descriptor.editor_custom_virtual_path = dct.get("EditorCustomVirtualPath", "")
        descriptor.enabled_by_default = dct.get("EnabledByDefault", False)
        descriptor.engine_version = dct.get("EngineVersion", "")
        descriptor.friendly_name = dct.get("FriendlyName", "")
        descriptor.localization_targets = dct.get("LocalizationTargets", [])
        descriptor.marketplace_url = dct.get("MarketplaceURL", "")
        descriptor.__modules = dct.get("Modules", [])
        descriptor.parent_plugin_name = dct.get("ParentPluginName", "")
        descriptor.__plugins = dct.get("Plugins", [])
        descriptor.post_build_steps = dct.get("PostBuildSteps")
        descriptor.pre_build_steps = dct.get("PreBuildSteps")
        descriptor.supported_programs = dct.get("SupportedPrograms", [])
        descriptor.supported_target_platforms = dct.get("SupportedTargetPlatforms", [])
        descriptor.support_url = dct.get("SupportURL", "")
        descriptor.version = dct.get("Version", 1)
        descriptor.version_name = dct.get("VersionName", "")

        if descriptor.is_valid():
            return descriptor
        return dct

    def to_dict(self) -> Dict[str, Any]:
        return {
            "CanContainContent": self.can_contain_content,
            "ExplicitlyLoaded": self.explicitly_loaded,
            "Installed": self.installed,
            "IsBetaVersion": self.is_beta_version,
            "IsExperimentalVersion": self.is_experimental_version,
            "IsHidden": self.is_hidden,
            "IsPluginExtension": self.is_plugin_extension,
            "RequiresBuildPlatform": self.requires_build_platform,
            "Category": self.category,
            "CreatedBy": self.created_by,
            "CreatedByURL": self.created_by_url,
            "Description": self.description,
            "DocsURL": self.docs_url,
            "EditorCustomVirtualPath": self.editor_custom_virtual_path,
            "EnabledByDefault": self.enabled_by_default,
            "EngineVersion": self.engine_version,
            "FriendlyName": self.friendly_name,
            "LocalizationTargets": self.localization_targets,
            "MarketplaceURL": self.marketplace_url,
            "Modules": list(self.modules),
            "ParentPluginName": self.parent_plugin_name,
            "Plugins": list(self.plugins),
            "PostBuildSteps": self.post_build_steps,
            "PreBuildSteps": self.pre_build_steps,
            "SupportedPrograms": self.supported_programs,
            "SupportedTargetPlatforms": self.supported_target_platforms,
            "SupportURL": self.support_url,
            "Version": self.version,
            "VersionName": self.version_name,
        }

    def is_valid(self) -> bool:
        return (
            self.friendly_name is not None
            and self.friendly_name != ""
            and self.version_name is not None
            and self.version_name != ""
        )

    def add_module(self, module: ModuleDescriptor) -> None:
        if not isinstance(module, ModuleDescriptor):
            raise NotImplementedError()
        self.__modules.append(module)


class PluginReferenceDescriptor(object):
    """Object wrapper representation of Unreal Plugin descriptor reference, equivalent to FPluginReferenceDescriptor.

    https://docs.unrealengine.com/en-US/API/Runtime/Projects/FPluginReferenceDescriptor/index.html
    """

    def __init__(self) -> None:
        """Initialize a new instance of PluginReferenceDescriptor."""
        self.enabled: bool = False
        self.blacklist_platforms: List[Any] = []
        self.blacklist_target_configurations: List[Any] = []
        self.blacklist_targets: List[Any] = []
        self.optional: bool = False
        self.description: str = ""
        self.marketplace_url: str = ""
        self.name: Optional[str] = None
        self.supported_target_platforms: List[Any] = []
        self.whitelist_platforms: List[Any] = []
        self.whitelist_target_configurations: List[Any] = []
        self.whitelist_targets: List[Any] = []

    def __repr__(self) -> str:
        """Python interpreter representation of PluginReferenceDescriptor."""
        return f"<PluginReferenceDescriptor {self.name}>"

    @staticmethod
    def to_object(
        dct: Dict[str, Any]
    ) -> Union[PluginReferenceDescriptor, Dict[str, Any]]:
        ref = PluginReferenceDescriptor()
        ref.enabled = dct.get("Enabled", False)
        ref.blacklist_platforms = dct.get("BlacklistPlatforms", [])
        ref.blacklist_target_configurations = dct.get(
            "BlacklistTargetConfigurations", []
        )
        ref.blacklist_targets = dct.get("BlacklistTargets", [])
        ref.optional = dct.get("Optional", False)
        ref.description = dct.get("Description", "")
        ref.marketplace_url = dct.get("MarketplaceURL", "")
        ref.name = dct.get("Name")
        ref.supported_target_platforms = dct.get("SupportedTargetPlatforms", [])
        ref.whitelist_platforms = dct.get("WhitelistPlatforms", [])
        ref.whitelist_target_configurations = dct.get(
            "WhitelistTargetConfigurations", []
        )
        ref.whitelist_targets = dct.get("WhitelistTargets", [])

        if ref.is_valid():
            return ref
        return dct

    def to_dict(self) -> Dict[str, Any]:
        return {
            "Enabled": self.enabled,
            "BlacklistPlatforms": self.blacklist_platforms,
            "BlacklistTargetConfigurations": self.blacklist_target_configurations,
            "BlacklistTargets": self.blacklist_targets,
            "Optional": self.optional,
            "Description": self.description,
            "MarketplaceURL": self.marketplace_url,
            "Name": self.name,
            "SupportedTargetPlatforms": self.supported_target_platforms,
            "WhitelistPlatforms": self.whitelist_platforms,
            "WhitelistTargetConfigurations": self.whitelist_target_configurations,
            "WhitelistTargets": self.whitelist_targets,
        }

    def is_valid(self) -> bool:
        return self.name is not None and self.name != ""


class UnrealPlugin(Buildable):
    """Object wrapper representation of an Unreal Engine plugin."""

    def __init__(self, plugin_file: str) -> None:
        if plugin_file is None:
            raise TypeError("UnrealPlugin plugin_file must not be None.")

        self.plugin_file: str = plugin_file
        self.__descriptor: Optional[PluginDescriptor] = None
        self.__name: Optional[str] = None
        self.__modules: Optional[Dict[str, ModuleDescriptor]] = None
        self.__plugin_refs: Optional[Dict[str, PluginReferenceDescriptor]] = None
        self.__code_templates: Optional[Dict[str, CodeTemplate]] = None

    def __repr__(self) -> str:
        """Python interpreter representation of UnrealPlugin."""
        return f"<UnrealPlugin at {self.plugin_file}>"

    @property
    def code_templates(self) -> Dict[str, CodeTemplate]:
        if self.__code_templates is None:
            self.__code_templates = {
                template.name: template
                for entry_point in entry_points().get("crazyhusk.code.listers", [])
                for template in entry_point.load()(self)
            }
        return self.__code_templates

    @property
    def descriptor(self) -> PluginDescriptor:
        if self.__descriptor is None:
            self.validate()

            with open(self.plugin_file, encoding="utf-8") as json_plugin_file:
                self.__descriptor = json.load(
                    json_plugin_file, object_hook=PluginDescriptor.to_object
                )

        return self.__descriptor

    @property
    def engine(self) -> Optional[UnrealEngine]:
        return None

    @property
    def modules(self) -> Dict[str, ModuleDescriptor]:
        if self.__modules is None:
            self.__modules = {
                module.name: module
                for module in self.descriptor.modules
                if isinstance(module, ModuleDescriptor) and module.name is not None
            }
        return self.__modules

    @property
    def name(self) -> str:
        if self.__name is None:
            self.__name = os.path.splitext(os.path.basename(self.plugin_file))[0]
        return self.__name

    @property
    def plugin_dir(self) -> str:
        """Directory path of this plugin."""
        return os.path.dirname(self.plugin_file)

    @property
    def plugin_refs(self) -> Dict[str, PluginReferenceDescriptor]:
        if self.__plugin_refs is None:
            self.__plugin_refs = {
                plugin.name: plugin
                for plugin in self.descriptor.plugins
                if isinstance(plugin, PluginReferenceDescriptor)
                and plugin.name is not None
            }
        return self.__plugin_refs

    @property
    def config_dir(self) -> str:
        """Directory path of this plugin's Config."""
        return os.path.join(self.plugin_dir, "Config")

    @property
    def content_dir(self) -> str:
        """Directory path of this plugin's Content."""
        return os.path.join(self.plugin_dir, "Content")

    # crazyhusk.code.listers
    @staticmethod
    def list_plugin_code_templates(plugin: UnrealPlugin) -> Iterable[CodeTemplate]:
        if isinstance(plugin, UnrealPlugin):
            for template_filename in glob.iglob(
                os.path.join(plugin.content_dir, "Editor", "Templates", "*.template")
            ):
                with open(
                    template_filename,
                    encoding="utf-8",
                ) as _template_file:
                    yield CodeTemplate(
                        os.path.basename(os.path.splitext(template_filename)[0]),
                        _template_file.read(),
                    )

    # crazyhusk.plugin.validators
    @staticmethod
    def plugin_file_exists(plugin: UnrealPlugin) -> None:
        """Raise exception if UnrealPlugin instance is not available on disk."""
        if not isinstance(plugin, UnrealPlugin):
            raise TypeError(
                f"Must provide an instance of crazyhusk.plugin.UnrealPlugin, got: {plugin!r}"
            )
        if not os.path.isfile(plugin.plugin_file):
            raise UnrealPluginError(
                f"Specified plugin descriptor file does not exist at {plugin.plugin_file}."
            )

    @staticmethod
    def valid_plugin_file_extension(plugin: UnrealPlugin) -> None:
        """Raise exception if UnrealPlugin instance does not have the correct file extension."""
        if not isinstance(plugin, UnrealPlugin):
            raise TypeError(
                f"Must provide an instance of crazyhusk.plugin.UnrealPlugin, got: {plugin!r}"
            )
        if not os.path.splitext(plugin.plugin_file)[-1] == ".uplugin":
            raise UnrealPluginError(f"Not a uplugin file: {plugin.plugin_file}")

    def get_build_command(
        self,
        target: Optional[str] = None,
        configuration: Optional[str] = None,
        platform: Optional[str] = None,
        *extra_switches: str,
        **extra_parameters: str,
    ) -> Iterable[str]:
        ...

    def is_buildable(self) -> bool:
        return False

    def unreal_path_to_file_path(
        self, unreal_path: str, ext: str = ".uasset"
    ) -> Optional[str]:
        """Convert an Unreal object path to a file path relative to this plugin."""
        path_split = unreal_path.split("/")
        if len(path_split) < 3:
            raise UnrealPluginError(f"Can't resolve Unreal path: {unreal_path}")

        mount = path_split[1]
        if mount == "Game":
            raise UnrealPluginError(
                f"Can't resolve Unreal path: {unreal_path} - could not resolve associated UnrealProject."
            )
        if mount == "Engine":
            raise UnrealPluginError(
                f"Can't resolve Unreal path: {unreal_path} - could not resolve associated UnrealEngine."
            )
        if mount == self.name:
            return os.path.join(self.content_dir, *path_split[2:]) + ext
        return None

    def unreal_path_from_file_path(self, file_path: str) -> Optional[str]:
        """Convert a file path to an appropriate Unreal object path for use with this plugin."""
        if (
            os.path.commonpath([os.path.realpath(file_path), self.content_dir])
            == self.content_dir
        ):
            sub_path = (
                os.path.splitext(os.path.realpath(file_path))[0]
                .split(self.content_dir)[1][1:]
                .replace(os.sep, "/")
            )
            return f"/{self.name}/{sub_path}"
        return None

    def validate(self) -> None:
        """Raise exceptions if this instance is misconfigured."""
        for entry_point in entry_points().get("crazyhusk.plugin.validators", []):
            entry_point.load()(self)



"""Wrapper objects for Unreal code modules."""

# Future Standard Library
from __future__ import annotations

# Standard Library
from typing import Any, Dict, List, Optional, Union

__all__ = ["ModuleDescriptor"]

HOST_TYPES = frozenset(
    {
        "Runtime",
        "RuntimeNoCommandlet",
        "RuntimeAndProgram",
        "CookedOnly",
        "UncookedOnly",
        "Developer",
        "DeveloperTool",
        "Editor",
        "EditorNoCommandlet",
        "EditorAndProgram",
        "Program",
        "ServerOnly",
        "ClientOnly",
        "ClientOnlyNoCommandlet",
    }
)

LOADING_PHASES = frozenset(
    {
        "EarliestPossible",
        "PostConfigInit",
        "PostSplashScreen",
        "PreEarlyLoadingScreen",
        "PreLoadingScreen",
        "PreDefault",
        "Default",
        "PostDefault",
        "PostEngineInit",
        "None",
    }
)


class ModuleDescriptor(object):
    """Object wrapper representation of Unreal code module, equivalent to FModuleDescriptor.

    https://docs.unrealengine.com/en-US/API/Runtime/Projects/FModuleDescriptor/index.html
    """

    def __init__(self) -> None:
        """Initialize a new instance of ModuleDescriptor."""
        self.name: Optional[str] = None
        self.host_type: Optional[str] = None
        self.loading_phase: Optional[str] = None
        self.additional_dependencies: List[Any] = []
        self.blacklist_platforms: List[Any] = []
        self.blacklist_programs: List[Any] = []
        self.blacklist_target_configurations: List[Any] = []
        self.blacklist_targets: List[Any] = []
        self.whitelist_platforms: List[Any] = []
        self.whitelist_programs: List[Any] = []
        self.whitelist_target_configurations: List[Any] = []
        self.whitelist_targets: List[Any] = []

    def __repr__(self) -> str:
        """Python interpreter representation of ModuleDescriptor."""
        return f"<ModuleDescriptor {self.name}>"

    def __hash__(self) -> int:
        """Provide a consistent identity."""
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        """Provide consistent equality operation."""
        if not isinstance(other, ModuleDescriptor):
            return NotImplemented
        return self.name == other.name

    def __ne__(self, other: object) -> bool:
        """Provide consistent inequality operation."""
        if not isinstance(other, ModuleDescriptor):
            return NotImplemented
        return self.name != other.name

    @staticmethod
    def to_object(dct: Dict[str, Any]) -> Union[ModuleDescriptor, Dict[str, Any]]:
        descriptor = ModuleDescriptor()
        descriptor.name = dct.get("Name")
        descriptor.host_type = dct.get("Type", "Runtime")
        descriptor.loading_phase = dct.get("LoadingPhase", "Default")
        descriptor.blacklist_platforms = dct.get("BlacklistPlatforms", [])
        descriptor.blacklist_programs = dct.get("BlacklistPrograms", [])
        descriptor.blacklist_target_configurations = dct.get(
            "BlacklistTargetConfigurations", []
        )
        descriptor.blacklist_targets = dct.get("BlacklistTargets", [])
        descriptor.whitelist_platforms = dct.get("WhitelistPlatforms", [])
        descriptor.whitelist_programs = dct.get("WhitelistPrograms", [])
        descriptor.whitelist_target_configurations = dct.get(
            "WhitelistTargetConfigurations", []
        )
        descriptor.whitelist_targets = dct.get("WhitelistTargets", [])

        if descriptor.is_valid():
            return descriptor
        return dct

    def to_dict(self) -> Dict[str, Any]:
        return {
            "Name": self.name,
            "Type": self.host_type,
            "LoadingPhase": self.loading_phase,
            "BlacklistPlatforms": self.blacklist_platforms,
            "BlacklistPrograms": self.blacklist_programs,
            "BlacklistTargetConfigurations": self.blacklist_target_configurations,
            "BlacklistTargets": self.blacklist_targets,
            "WhitelistPlatforms": self.whitelist_platforms,
            "WhitelistPrograms": self.whitelist_programs,
            "WhitelistTargetConfigurations": self.whitelist_target_configurations,
            "WhitelistTargets": self.whitelist_targets,
        }

    def is_valid(self) -> bool:
        return (
            self.name is not None
            and self.name != ""
            and self.host_type in HOST_TYPES
            and self.loading_phase in LOADING_PHASES
        )



# Standard Library
import json
import os
import types
from typing import Any, Dict, Iterable, Optional, Type

# Third Party
import pytest

# CrazyHusk
from crazyhusk import code, module, plugin


def test_plugin_descriptor_init(
    empty_name_plugin_descriptor: plugin.PluginDescriptor,
) -> None:
    assert not empty_name_plugin_descriptor.can_contain_content
    assert not empty_name_plugin_descriptor.explicitly_loaded
    assert not empty_name_plugin_descriptor.installed
    assert not empty_name_plugin_descriptor.is_beta_version
    assert not empty_name_plugin_descriptor.is_experimental_version
    assert not empty_name_plugin_descriptor.is_hidden
    assert not empty_name_plugin_descriptor.is_plugin_extension
    assert not empty_name_plugin_descriptor.requires_build_platform
    assert empty_name_plugin_descriptor.category == ""
    assert empty_name_plugin_descriptor.created_by == ""
    assert empty_name_plugin_descriptor.created_by_url == ""
    assert empty_name_plugin_descriptor.description == ""
    assert empty_name_plugin_descriptor.docs_url == ""
    assert empty_name_plugin_descriptor.editor_custom_virtual_path == ""
    assert not empty_name_plugin_descriptor.enabled_by_default
    assert empty_name_plugin_descriptor.engine_version == ""
    assert empty_name_plugin_descriptor.friendly_name == ""
    assert isinstance(empty_name_plugin_descriptor.localization_targets, list)
    assert len(empty_name_plugin_descriptor.localization_targets) == 0
    assert empty_name_plugin_descriptor.marketplace_url == ""
    assert isinstance(empty_name_plugin_descriptor.modules, types.GeneratorType)
    assert empty_name_plugin_descriptor.parent_plugin_name == ""
    assert isinstance(empty_name_plugin_descriptor.plugins, types.GeneratorType)
    assert empty_name_plugin_descriptor.post_build_steps is None
    assert empty_name_plugin_descriptor.pre_build_steps is None
    assert isinstance(empty_name_plugin_descriptor.supported_programs, list)
    assert len(empty_name_plugin_descriptor.supported_programs) == 0
    assert isinstance(empty_name_plugin_descriptor.supported_target_platforms, list)
    assert len(empty_name_plugin_descriptor.supported_target_platforms) == 0
    assert empty_name_plugin_descriptor.support_url == ""
    assert empty_name_plugin_descriptor.version == 1
    assert empty_name_plugin_descriptor.version_name == ""


def test_plugin_descriptor_repr(
    empty_name_plugin_descriptor: plugin.PluginDescriptor,
) -> None:
    assert repr(empty_name_plugin_descriptor) == "<PluginDescriptor  version >"


@pytest.mark.parametrize(
    "plugin_descriptor_fixture,expected_type",
    [
        ("null_name_plugin_descriptor", dict),
        ("empty_name_plugin_descriptor", dict),
        ("null_version_plugin_descriptor", dict),
        ("empty_version_plugin_descriptor", dict),
        ("basic_plugin_descriptor", plugin.PluginDescriptor),
        ("basic_plugin_descriptor_withmodule", plugin.PluginDescriptor),
    ],
)
def test_plugin_descriptor_to_object(
    plugin_descriptor_fixture: str, expected_type: Type[Any], request: Any
) -> None:
    plugin_descriptor = request.getfixturevalue(plugin_descriptor_fixture)
    dct = plugin_descriptor.to_dict()
    assert isinstance(plugin.PluginDescriptor.to_object(dct), expected_type)


def test_plugin_descriptor_to_object_module_dict(
    basic_plugin_descriptor_withmodule_dict: Iterable[Dict[str, Any]],
) -> None:
    _plugin = plugin.PluginDescriptor.to_object(basic_plugin_descriptor_withmodule_dict)
    assert isinstance(_plugin, plugin.PluginDescriptor)
    for _module in _plugin.modules:
        assert isinstance(_module, module.ModuleDescriptor)


@pytest.mark.parametrize(
    "plugin_descriptor_fixture,expected_type",
    [
        ("null_name_plugin_descriptor", None),
        ("empty_name_plugin_descriptor", None),
        ("null_version_plugin_descriptor", None),
        ("empty_version_plugin_descriptor", None),
        ("basic_plugin_descriptor", None),
        ("basic_plugin_descriptor_withmodule", module.ModuleDescriptor),
    ],
)
def test_plugin_descriptor_modules(
    plugin_descriptor_fixture: str, expected_type: Optional[Type[Any]], request: Any
) -> None:
    plugin_descriptor = request.getfixturevalue(plugin_descriptor_fixture)
    for _module in plugin_descriptor.modules:
        if expected_type is not None:
            assert isinstance(_module, expected_type)


def test_plugin_descriptor_add_module(
    basic_plugin_descriptor: plugin.PluginDescriptor,
    default_valid_module_descriptor: module.ModuleDescriptor,
) -> None:
    with pytest.raises(NotImplementedError):
        assert basic_plugin_descriptor.add_module({}) is None
    assert basic_plugin_descriptor.add_module(default_valid_module_descriptor) is None


# PluginReferenceDescriptor tests


def test_plugin_reference_descriptor_init(
    null_plugin_reference_descriptor: plugin.PluginReferenceDescriptor,
) -> None:
    assert not null_plugin_reference_descriptor.enabled
    assert isinstance(null_plugin_reference_descriptor.blacklist_platforms, list)
    assert len(null_plugin_reference_descriptor.blacklist_platforms) == 0
    assert isinstance(
        null_plugin_reference_descriptor.blacklist_target_configurations, list
    )
    assert len(null_plugin_reference_descriptor.blacklist_target_configurations) == 0
    assert isinstance(null_plugin_reference_descriptor.blacklist_targets, list)
    assert len(null_plugin_reference_descriptor.blacklist_targets) == 0
    assert not null_plugin_reference_descriptor.optional
    assert null_plugin_reference_descriptor.description == ""
    assert null_plugin_reference_descriptor.marketplace_url == ""
    assert null_plugin_reference_descriptor.name is None
    assert isinstance(null_plugin_reference_descriptor.supported_target_platforms, list)
    assert len(null_plugin_reference_descriptor.supported_target_platforms) == 0
    assert isinstance(null_plugin_reference_descriptor.whitelist_platforms, list)
    assert len(null_plugin_reference_descriptor.whitelist_platforms) == 0
    assert isinstance(
        null_plugin_reference_descriptor.whitelist_target_configurations, list
    )
    assert len(null_plugin_reference_descriptor.whitelist_target_configurations) == 0
    assert isinstance(null_plugin_reference_descriptor.whitelist_targets, list)
    assert len(null_plugin_reference_descriptor.whitelist_targets) == 0


def test_plugin_reference_descriptor_repr(
    null_plugin_reference_descriptor: plugin.PluginReferenceDescriptor,
) -> None:
    assert repr(null_plugin_reference_descriptor) == "<PluginReferenceDescriptor None>"


@pytest.mark.parametrize(
    "plugin_reference_descriptor_fixture,expected_type",
    [
        ("null_plugin_reference_descriptor", dict),
        ("empty_plugin_reference_descriptor", dict),
        ("basic_plugin_reference_descriptor", plugin.PluginReferenceDescriptor),
    ],
)
def test_plugin_reference_descriptor_to_object(
    plugin_reference_descriptor_fixture: str, expected_type: Type[Any], request: Any
) -> None:
    plugin_reference_descriptor = request.getfixturevalue(
        plugin_reference_descriptor_fixture
    )
    dct = plugin_reference_descriptor.to_dict()
    assert isinstance(plugin.PluginReferenceDescriptor.to_object(dct), expected_type)


# UnrealPlugin tests


@pytest.mark.parametrize(
    "plugin_file,raises",
    [(None, TypeError), ("", None)],
)
def test_unreal_plugin_init(
    plugin_file: Optional[str], raises: Optional[Type[BaseException]]
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert plugin.UnrealPlugin(plugin_file)
    else:
        _plugin = plugin.UnrealPlugin(plugin_file)
        assert _plugin.plugin_file == plugin_file


def test_unreal_plugin_repr(empty_unreal_plugin: plugin.UnrealPlugin) -> None:
    assert repr(empty_unreal_plugin) == "<UnrealPlugin at >"


@pytest.mark.parametrize(
    "unreal_plugin_fixture,plugin_dir,name,config_dir,content_dir",
    [
        ("empty_unreal_plugin", "", "", "Config", "Content"),
        ("local_dir_unreal_plugin", "", ".", "Config", "Content"),
        (
            "invalid_file_unreal_plugin",
            ".",
            "test",
            os.path.join(".", "Config"),
            os.path.join(".", "Content"),
        ),
    ],
)
def test_unreal_plugin_properties(
    unreal_plugin_fixture: str,
    plugin_dir: str,
    name: str,
    config_dir: str,
    content_dir: str,
    request: Any,
) -> None:
    unreal_plugin = request.getfixturevalue(unreal_plugin_fixture)
    assert unreal_plugin.plugin_dir == plugin_dir
    assert unreal_plugin.name == name
    assert unreal_plugin.config_dir == config_dir
    assert unreal_plugin.content_dir == content_dir


@pytest.mark.parametrize(
    "unreal_plugin_fixture,raises",
    [
        ("empty_unreal_plugin", plugin.UnrealPluginError),
        ("local_dir_unreal_plugin", plugin.UnrealPluginError),
        ("invalid_file_unreal_plugin", plugin.UnrealPluginError),
        ("empty_file_content_unreal_plugin", None),
    ],
)
def test_unreal_plugin_file_exists(
    unreal_plugin_fixture: str, raises: Optional[Type[BaseException]], request: Any
) -> None:
    unreal_plugin = request.getfixturevalue(unreal_plugin_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert plugin.UnrealPlugin.plugin_file_exists(unreal_plugin) is None
    else:
        assert plugin.UnrealPlugin.plugin_file_exists(unreal_plugin) is None


def test_unreal_plugin_file_exists_types() -> None:
    with pytest.raises(TypeError):
        assert plugin.UnrealPlugin.plugin_file_exists(None)


@pytest.mark.parametrize(
    "unreal_plugin_fixture,raises",
    [
        ("empty_unreal_plugin", plugin.UnrealPluginError),
        ("local_dir_unreal_plugin", plugin.UnrealPluginError),
        ("invalid_file_unreal_plugin", plugin.UnrealPluginError),
        ("empty_file_content_unreal_plugin", None),
    ],
)
def test_unreal_plugin_file_extension(
    unreal_plugin_fixture: str, raises: Optional[Type[BaseException]], request: Any
) -> None:
    unreal_plugin = request.getfixturevalue(unreal_plugin_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert (
                plugin.UnrealPlugin.valid_plugin_file_extension(unreal_plugin) is None
            )
    else:
        assert plugin.UnrealPlugin.valid_plugin_file_extension(unreal_plugin) is None


def test_unreal_plugin_file_extension_types() -> None:
    with pytest.raises(TypeError):
        assert plugin.UnrealPlugin.valid_plugin_file_extension(None)


@pytest.mark.parametrize(
    "unreal_path,raises,expected",
    [
        ("", plugin.UnrealPluginError, None),
        ("/", plugin.UnrealPluginError, None),
        ("/Game", plugin.UnrealPluginError, None),
        ("/Game/whatever", plugin.UnrealPluginError, None),
        ("/Engine", plugin.UnrealPluginError, None),
        ("/Engine/whatever", plugin.UnrealPluginError, None),
        ("/Invalid", plugin.UnrealPluginError, None),
        ("/Invalid/whatever", None, None),
        ("/test", plugin.UnrealPluginError, None),
        ("/test/whatever", None, os.path.join(".", "Content", "whatever.uasset")),
    ],
)
def test_unreal_path_to_file_path(
    unreal_path: str,
    raises: Optional[Type[BaseException]],
    expected: Optional[str],
    invalid_file_unreal_plugin: plugin.UnrealPlugin,
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert (
                invalid_file_unreal_plugin.unreal_path_to_file_path(unreal_path)
                == expected
            )
    else:
        assert (
            invalid_file_unreal_plugin.unreal_path_to_file_path(unreal_path) == expected
        )


@pytest.mark.parametrize(
    "file_path,raises,expected",
    [
        ("", None, None),
        ("/", None, None),
        (".", None, None),
        (os.path.realpath("."), None, None),
        (os.path.realpath("./Content"), None, "/test/"),
        (
            os.path.realpath(os.path.join(".", "Content", "whatever.uasset")),
            None,
            "/test/whatever",
        ),
    ],
)
def test_unreal_path_from_file_path(
    file_path: str,
    raises: Optional[Type[BaseException]],
    expected: Optional[str],
    invalid_file_unreal_plugin_realpath: plugin.UnrealPlugin,
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert (
                invalid_file_unreal_plugin_realpath.unreal_path_from_file_path(
                    file_path
                )
                == expected
            )
    else:
        assert (
            invalid_file_unreal_plugin_realpath.unreal_path_from_file_path(file_path)
            == expected
        )


def test_unreal_plugin_code_templates(
    empty_file_content_unreal_plugin: plugin.UnrealPlugin,
) -> None:
    plugin.entry_points = lambda: {}
    for _name, _plugin in empty_file_content_unreal_plugin.code_templates.items():
        assert isinstance(_name, str)
        assert isinstance(_plugin, code.CodeTemplate)


def test_unreal_plugin_properties_types(
    empty_file_content_unreal_plugin: plugin.UnrealPlugin,
    basic_unreal_plugin: plugin.UnrealPlugin,
) -> None:
    plugin.entry_points = lambda: {}
    with pytest.raises(json.decoder.JSONDecodeError):
        assert empty_file_content_unreal_plugin.descriptor is None

    for _name, _module in basic_unreal_plugin.modules.items():
        assert isinstance(_name, str)
        assert isinstance(_module, module.ModuleDescriptor)

    for _name, _ref in basic_unreal_plugin.plugin_refs.items():
        assert isinstance(_name, str)
        assert isinstance(_ref, plugin.PluginReferenceDescriptor)


def test_unreal_plugin_list_plugin_code_templates(
    basic_unreal_plugin: plugin.UnrealPlugin,
) -> None:
    assert (
        len(list(plugin.UnrealPlugin.list_plugin_code_templates(basic_unreal_plugin)))
        == 0
    )


def test_unreal_plugin_validate(
    basic_unreal_plugin: plugin.UnrealPlugin, monkeypatch: Any, test_entry_point: Any
) -> None:
    monkeypatch.setattr(plugin, "entry_points", lambda: {})
    assert basic_unreal_plugin.validate() is None

    monkeypatch.setattr(
        plugin,
        "entry_points",
        lambda: {"crazyhusk.plugin.validators": [test_entry_point]},
    )
    assert basic_unreal_plugin.validate() is None



# Standard Library
import json
from typing import Any, Dict, List, Optional, Type
from xml.etree.ElementTree import Element

# Third Party
import pytest
from pyexpat import ExpatError

# CrazyHusk
from crazyhusk import reports


@pytest.mark.parametrize(
    "timestamp,raises,expected",
    [
        (None, TypeError, None),
        (123, TypeError, None),
        ("", ValueError, None),
        ("123456", ValueError, None),
        ("1234.56.78-12.34.56", ValueError, None),
        ("2022.02.24-14.02.39", None, "2022-02-24T14:02:39"),
    ],
)
def test_report_timestamp_to_iso8601_timestamp(
    timestamp: Any, raises: Optional[Type[BaseException]], expected: Optional[str]
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert reports.report_timestamp_to_iso8601_timestamp(timestamp) is None
    else:
        assert reports.report_timestamp_to_iso8601_timestamp(timestamp) == expected


@pytest.mark.parametrize(
    "entry,raises",
    [
        (None, AttributeError),
        ("", AttributeError),
        ({}, None),
        ({"timestamp": None}, None),
        ({"timestamp": ""}, ValueError),
        ({"timestamp": "123456"}, ValueError),
        ({"timestamp": "1234.56.78-12.34.56"}, ValueError),
        ({"timestamp": "2022.02.24-14.02.39"}, None),
    ],
)
def test_report_entry_to_entry_xml(
    entry: Dict[str, Any], raises: Optional[Type[BaseException]]
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert isinstance(reports.report_entry_to_entry_xml(entry), Element)
    else:
        assert isinstance(reports.report_entry_to_entry_xml(entry), Element)


@pytest.mark.parametrize(
    "test,raises",
    [
        (None, AttributeError),
        ("", AttributeError),
        ({}, None),
        ({"entries": None}, TypeError),
        ({"entries": ""}, None),
        ({"entries": "randomstring"}, AttributeError),
        ({"entries": []}, None),
        ({"entries": [{}]}, None),
        ({"entries": [{"timestamp": None}]}, None),
        ({"entries": [{"timestamp": ""}]}, ValueError),
        ({"entries": [{"timestamp": "2022.02.24-14.02.39"}]}, None),
    ],
)
def test_report_test_to_testcase_xml(
    test: Dict[str, Any], raises: Optional[Type[BaseException]]
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert isinstance(reports.report_test_to_testcase_xml(test), Element)
    else:
        assert isinstance(reports.report_test_to_testcase_xml(test), Element)


@pytest.mark.parametrize(
    "report,raises",
    [
        (None, AttributeError),
        ("", AttributeError),
        ({}, None),
        ({"reportCreatedOn": None}, None),
        ({"reportCreatedOn": ""}, ValueError),
        ({"reportCreatedOn": "123456"}, ValueError),
        ({"reportCreatedOn": "1234.56.78-12.34.56"}, ValueError),
        ({"reportCreatedOn": "2022.02.24-14.02.39"}, None),
        ({"tests": None}, TypeError),
        ({"tests": ""}, None),
        ({"tests": "randomstring"}, AttributeError),
        ({"tests": []}, None),
        ({"tests": [{}]}, None),
    ],
)
def test_report_object_to_testsuite_xml(
    report: Dict[str, Any], raises: Optional[Type[BaseException]]
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert isinstance(reports.report_object_to_testsuite_xml(report), Element)
    else:
        assert isinstance(reports.report_object_to_testsuite_xml(report), Element)


@pytest.mark.parametrize(
    "report_file_fixture,raises",
    [
        ("null_report_file", TypeError),
        ("empty_filename_report_file", ValueError),
        ("non_json_report_file", ValueError),
        ("empty_json_report_file", json.JSONDecodeError),
        ("list_json_report_file", ValueError),
    ],
)
def test_json_report_to_dict(
    report_file_fixture: str, raises: Optional[Type[BaseException]], request: Any
) -> None:
    report_file = request.getfixturevalue(report_file_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert isinstance(reports.json_report_to_dict(report_file), dict)
    else:
        assert isinstance(reports.json_report_to_dict(report_file), dict)


@pytest.mark.parametrize(
    "report_file_fixture,test_suites_fixture,raises",
    [
        ("null_report_file", "null_test_suites", TypeError),
        ("empty_filename_report_file", "emptystring_test_suites", ValueError),
        ("xml_filename_report_file", "emptystring_test_suites", AttributeError),
        ("xml_filename_mkdirs_report_file", "emptystring_test_suites", AttributeError),
        ("xml_filename_report_file", "empty_element_test_suites", ExpatError),
        ("xml_filename_report_file", "basic_element_test_suites", None),
    ],
)
def test_write_junit_xml_report(
    report_file_fixture: str,
    test_suites_fixture: str,
    raises: Optional[Type[BaseException]],
    request: Any,
) -> None:
    report_file = request.getfixturevalue(report_file_fixture)
    test_suites = request.getfixturevalue(test_suites_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert reports.write_junit_xml_report(report_file, test_suites) is None
    else:
        assert reports.write_junit_xml_report(report_file, test_suites) is None


@pytest.mark.parametrize(
    "report_file_fixture,json_file_fixtures,raises",
    [
        ("null_report_file", ["empty_dict_json_report_file"], None),
        ("empty_filename_report_file", ["empty_dict_json_report_file"], ValueError),
        ("xml_filename_report_file", ["empty_dict_json_report_file"], None),
    ],
)
def test_json_reports_to_junit_xml(
    report_file_fixture: str,
    json_file_fixtures: List[str],
    raises: Optional[Type[BaseException]],
    request: Any,
) -> None:
    report_file = request.getfixturevalue(report_file_fixture)
    json_reports = [
        request.getfixturevalue(json_report) for json_report in json_file_fixtures
    ]
    if raises is not None:
        with pytest.raises(raises):
            assert reports.json_reports_to_junit_xml(report_file, json_reports) is None
    else:
        assert reports.json_reports_to_junit_xml(report_file, json_reports) is None



"""Object wrappers for Unreal projects."""

# Future Standard Library
from __future__ import annotations

# Standard Library
import glob
import json
import os
from copy import deepcopy
from typing import Any, Dict, Iterable, List, Optional, Union

try:
    # Standard Library
    from importlib.metadata import entry_points  # type:ignore
except ImportError:
    # Third Party
    from importlib_metadata import entry_points  # type:ignore

# CrazyHusk
from crazyhusk.build import Buildable
from crazyhusk.code import CodeTemplate
from crazyhusk.config import CONFIG_CATEGORIES, UnrealConfigParser
from crazyhusk.engine import UnrealEngine
from crazyhusk.module import ModuleDescriptor
from crazyhusk.plugin import PluginReferenceDescriptor, UnrealPlugin

__all__ = ["UnrealProject"]


class UnrealProjectError(Exception):
    """Custom exception representing errors encountered with UnrealProject."""


class ProjectDescriptor(object):
    """Object wrapper representation of a uproject file, equivalent to serialization method used with FProjectDescriptor.

    https://docs.unrealengine.com/en-US/API/Runtime/Projects/FProjectDescriptor/index.html
    """

    def __init__(self) -> None:
        self.engine_association: Optional[str] = None
        self.category: str = ""
        self.description: str = ""
        self.disable_engine_plugins_by_default: bool = False
        self.is_enterprise_project: bool = False
        self.epic_sample_name_hash: Optional[str] = None
        self.post_build_steps: Optional[List[Any]] = None
        self.pre_build_steps: Optional[List[Any]] = None
        self.target_platforms: List[Any] = []
        self.__plugins: List[Dict[str, Any]] = []
        self.__modules: List[Dict[str, Any]] = []

    def __repr__(self) -> str:
        """Python interpreter representation of ProjectDescriptor."""
        return f"<ProjectDescriptor {self.description}>"

    @property
    def modules(self) -> Iterable[Union[ModuleDescriptor, Dict[str, Any]]]:
        for module in self.__modules:
            yield ModuleDescriptor.to_object(module)

    @property
    def plugins(self) -> Iterable[Union[PluginReferenceDescriptor, Dict[str, Any]]]:
        for plugin in self.__plugins:
            yield PluginReferenceDescriptor.to_object(plugin)

    @staticmethod
    def to_object(dct: Dict[str, Any]) -> Union[ProjectDescriptor, Dict[str, Any]]:
        descriptor = ProjectDescriptor()
        descriptor.engine_association = dct.get("EngineAssociation")
        descriptor.category = dct.get("Category", "")
        descriptor.description = dct.get("Description", "")
        descriptor.disable_engine_plugins_by_default = dct.get(
            "DisableEnginePluginsByDefault", False
        )
        descriptor.is_enterprise_project = dct.get("IsEnterpriseProject", False)
        descriptor.epic_sample_name_hash = dct.get("EpicSampleNameHash")
        descriptor.post_build_steps = dct.get("PostBuildSteps")
        descriptor.pre_build_steps = dct.get("PreBuildSteps")
        descriptor.target_platforms = dct.get("TargetPlatforms", [])
        descriptor.__plugins = dct.get("Plugins", [])
        descriptor.__modules = dct.get("Modules", [])

        if descriptor.is_valid():
            return descriptor
        return dct

    def to_dict(self) -> Dict[str, Any]:
        return {
            "EngineAssociation": self.engine_association,
            "Category": self.category,
            "Description": self.description,
            "DisableEnginePluginsByDefault": self.disable_engine_plugins_by_default,
            "IsEnterpriseProject": self.is_enterprise_project,
            "EpicSampleNameHash": self.epic_sample_name_hash,
            "PostBuildSteps": self.post_build_steps,
            "PreBuildSteps": self.pre_build_steps,
            "TargetPlatforms": self.target_platforms,
            "Modules": list(self.modules),
            "Plugins": list(self.plugins),
        }

    def is_valid(self) -> bool:
        return self.engine_association is not None

    def add_module(self, module: ModuleDescriptor) -> None:
        if not isinstance(module, ModuleDescriptor):
            raise NotImplementedError()
        self.__modules.append(module.to_dict())


class UnrealProject(Buildable):
    """Object wrapper representation of an Unreal Engine project."""

    def __init__(self, project_file: str) -> None:
        self.project_file: str = project_file
        self.name: str = os.path.splitext(os.path.basename(project_file))[0]

        self.__descriptor: Optional[ProjectDescriptor] = None
        self.__engine: Optional[UnrealEngine] = None
        self.__modules: Optional[Dict[str, ModuleDescriptor]] = None
        self.__plugins: Optional[Dict[str, UnrealPlugin]] = None
        self.__code_templates: Optional[Dict[str, CodeTemplate]] = None

    def __repr__(self) -> str:
        """Python interpreter representation."""
        return f"<UnrealProject {self.name} at {self.project_file}>"

    @property
    def code_templates(self) -> Dict[str, CodeTemplate]:
        if self.__code_templates is None:
            self.__code_templates = {}
            items = []  # type: List[Union[UnrealProject,UnrealEngine,UnrealPlugin]]
            if self.engine is not None:
                items.append(self.engine)
                if self.engine.plugins is not None and len(self.engine.plugins):
                    items.append(*self.engine.plugins.values())
            items.append(self)
            if self.plugins is not None and len(self.plugins):
                items.append(*self.plugins.values())
            for entry_point in entry_points().get("crazyhusk.code.listers", []):
                for item in items:
                    for template in entry_point.load()(item):
                        self.__code_templates[template.name] = template
        return self.__code_templates

    @property
    def descriptor(self) -> Optional[ProjectDescriptor]:
        if self.__descriptor is None:
            self.validate()

            with open(self.project_file, encoding="utf-8") as json_project_file:
                self.__descriptor = json.load(
                    json_project_file, object_hook=ProjectDescriptor.to_object
                )

        return self.__descriptor

    @property
    def engine(self) -> Optional[UnrealEngine]:
        if self.__engine is None:
            if self.descriptor is not None and self.descriptor.engine_association == "":
                self.__engine = UnrealEngine(
                    os.path.realpath(os.path.join(self.project_file, "..", "..")), ""
                )
            elif (
                self.descriptor is not None
                and self.descriptor.engine_association is not None
            ):
                self.__engine = UnrealEngine.find_engine(
                    self.descriptor.engine_association
                )
        return self.__engine

    @engine.setter
    def engine(self, new_engine: Union[UnrealEngine, str]) -> None:
        if not isinstance(new_engine, UnrealEngine):
            self.__engine = UnrealEngine.find_engine(new_engine)
        else:
            self.__engine = new_engine

    @property
    def project_dir(self) -> str:
        """Get the base directory for .uproject file."""
        return os.path.dirname(self.project_file)

    @property
    def config_dir(self) -> str:
        """Get the project's Config directory."""
        return os.path.join(self.project_dir, "Config")

    @property
    def content_dir(self) -> str:
        """Get the project's Content directory."""
        return os.path.join(self.project_dir, "Content")

    @property
    def plugins_dir(self) -> str:
        """Get the project's Plugins directory."""
        return os.path.join(self.project_dir, "Plugins")

    @property
    def modules(self) -> Optional[Dict[str, ModuleDescriptor]]:
        if self.__modules is None:
            if self.descriptor is not None:
                self.__modules = {
                    module.name: module
                    for module in self.descriptor.modules
                    if isinstance(module, ModuleDescriptor) and module.name is not None
                }
        return self.__modules

    @property
    def plugins(self) -> Optional[Dict[str, UnrealPlugin]]:
        if self.__plugins is None:
            if self.engine is None:
                self.__plugins = {}
            else:
                self.__plugins = deepcopy(self.engine.plugins)

            for _root, _dirs, _files in os.walk(self.plugins_dir):
                for _file in _files:
                    if os.path.splitext(_file)[-1] == ".uplugin":
                        plugin = UnrealPlugin(os.path.join(_root, _file))
                        if self.__plugins is not None and plugin.name is not None:
                            self.__plugins[plugin.name] = plugin
                        break
        return self.__plugins

    @property
    def saved_dir(self) -> str:
        """Get the project's Saved directory."""
        return os.path.join(self.project_dir, "Saved")

    @property
    def reports_dir(self) -> str:
        """Get the project's default Reports directory."""
        return os.path.join(self.project_dir, "Saved", "Reports")

    # crazyhusk.code.listers
    @staticmethod
    def list_project_code_templates(project: UnrealProject) -> Iterable[CodeTemplate]:
        if isinstance(project, UnrealProject):
            for template_filename in glob.iglob(
                os.path.join(project.content_dir, "Editor", "Templates", "*.template")
            ):
                with open(
                    template_filename,
                    encoding="utf-8",
                ) as _template_file:
                    yield CodeTemplate(
                        os.path.basename(os.path.splitext(template_filename)[0]),
                        _template_file.read(),
                    )

    # crazyhusk.project.validators
    @staticmethod
    def project_file_exists(project: UnrealProject) -> None:
        """Raise exception if UnrealProject instance is not available on disk."""
        if not isinstance(project, UnrealProject):
            raise TypeError(
                f"Must provide an instance of crazyhusk.project.UnrealProject, got: {project!r}"
            )
        if not os.path.isfile(project.project_file):
            raise UnrealProjectError(
                f"Specified project file does not exist at {project.project_file}."
            )

    @staticmethod
    def valid_project_file_extension(project: UnrealProject) -> None:
        """Raise exception if UnrealProject instance does not have the correct file extension."""
        if not isinstance(project, UnrealProject):
            raise TypeError(
                f"Must provide an instance of crazyhusk.project.UnrealProject, got: {project!r}"
            )
        if not os.path.splitext(project.project_file)[-1] == ".uproject":
            raise UnrealProjectError(f"Not a uproject file: {project.project_file}")

    def config(
        self, config_category: Optional[str] = None, platform: Optional[str] = None
    ) -> UnrealConfigParser:
        """Create a configuration object associated with this project by category and platform."""
        _config = UnrealConfigParser()
        if isinstance(self.engine, UnrealEngine):
            self.engine.validate()
            _config.read(self.engine.config_files(config_category, platform))
        _config.read(self.config_files(config_category, platform))
        return _config

    def config_files(
        self, config_category: Optional[str] = None, platform: Optional[str] = None
    ) -> Iterable[str]:
        """Iterate configuration file paths associated with this project by category and platform."""
        if config_category in CONFIG_CATEGORIES:
            yield os.path.join(self.config_dir, f"Default{config_category}.ini")
            if platform is not None:
                yield os.path.join(
                    self.config_dir, platform, f"{platform}{config_category}.ini"
                )

    def get_build_command(
        self,
        target: Optional[str] = None,
        configuration: Optional[str] = None,
        platform: Optional[str] = None,
        *extra_switches: str,
        **extra_parameters: str,
    ) -> Iterable[str]:
        if self.engine is None:
            raise UnrealProjectError(
                f"Could not resolve an associated UnrealEngine for project: {self!r}"
            )
        ubt_path = self.engine.executable_path("UnrealBuildTool")
        if ubt_path is None:
            raise UnrealProjectError(
                f"Could not resolve a valid path to UnrealBuildTool for project: {self!r}"
            )
        yield ubt_path
        yield configuration or ""
        yield platform or ""
        switches = {"Progress", "WaitMutex", "NoHotReloadFromIDE"}
        switches.update(*extra_switches)
        parameters: Dict[str, str] = {
            "Project": self.project_file,
            "TargetType": target or "",
        }
        parameters.update(**extra_parameters)
        for arg in UnrealEngine.format_commandline_options(*switches, **parameters):
            yield arg

    def is_buildable(self) -> bool:
        return self.engine is not None  # TODO: check for .Target.cs files

    def list_tests(
        self, editor: bool = True, *extra_switches: str, **extra_parameters: str
    ) -> int:
        """List available automation tests for this project."""
        switches = {
            "buildmachine",
            "unattended",
            "nopause",
            "nullrhi",
            "stdout",
            "nosplash",
        }

        if editor:
            switches.add("editortest")
        else:
            switches.add("-game")

        switches.update(*extra_switches)

        params = {
            "ExecCmds": "Automation List; quit",
            "TestExit": "Automation Test Queue Empty",
        }
        params.update(extra_parameters)

        if self.engine is not None:
            editor_cmd_path = self.engine.executable_path("UE4Editor-Cmd")
            if editor_cmd_path is not None:
                with self.engine:
                    return self.engine.run(
                        editor_cmd_path,
                        f'"{self.project_file}"',
                        *UnrealEngine.format_commandline_options(*switches, **params),
                    )
        return -1

    def run_tests(
        self,
        tests: List[str],
        report_path: Optional[str] = None,
        editor: bool = True,
        *extra_switches: str,
        **extra_parameters: str,
    ) -> int:
        """Run named automation tests for this project."""
        if report_path is None:
            report_path = self.reports_dir

        switches = {
            "buildmachine",
            "unattended",
            "nopause",
            "nullrhi",
            "stdout",
            "nosplash",
        }

        if editor:
            switches.add("editortest")
        else:
            switches.add("-game")

        switches.update(*extra_switches)

        params = {
            "ExecCmds": "Automation RunTests " + "+".join(tests) + "; quit",
            "TestExit": "Automation Test Queue Empty",
            "ReportOutputPath": report_path,
        }
        params.update(extra_parameters)

        if self.engine is not None:
            editor_cmd_path = self.engine.executable_path("UE4Editor-Cmd")
            if editor_cmd_path is not None:
                with self.engine:
                    return self.engine.run(
                        editor_cmd_path,
                        f'"{self.project_file}"',
                        *UnrealEngine.format_commandline_options(*switches, **params),
                    )
        return -1

    def unreal_path_to_file_path(
        self, unreal_path: str, ext: str = ".uasset"
    ) -> Optional[str]:
        """Convert an Unreal object path to a file path relative to this project."""
        path_split = unreal_path.split("/")
        if len(path_split) < 3:
            raise UnrealProjectError(f"Can't resolve Unreal path: {unreal_path}")

        mount = path_split[1]
        if mount == "Game":
            return os.path.join(self.content_dir, *path_split[2:]) + ext

        if mount == "Engine":
            if not isinstance(self.engine, UnrealEngine):
                raise UnrealProjectError(
                    f"Can't resolve Unreal path: {unreal_path} - could not resolve associated UnrealEngine."
                )
            return os.path.join(self.engine.content_dir, *path_split[2:]) + ext

        if self.plugins is not None and mount in self.plugins:
            return self.plugins[mount].unreal_path_to_file_path(unreal_path, ext)

        raise UnrealProjectError(
            f"Can't resolve Unreal path: {unreal_path} - could not find plugin or feature pack mount {mount}."
        )

    def unreal_path_from_file_path(self, file_path: str) -> Optional[str]:
        """Convert a file path to an appropriate Unreal object path for use with this project."""
        if (
            os.path.commonpath([os.path.realpath(file_path), self.content_dir])
            == self.content_dir
        ):
            sub_path = (
                os.path.splitext(os.path.realpath(file_path))[0]
                .split(self.content_dir)[1][1:]
                .replace(os.sep, "/")
            )
            return f"/Game/{sub_path}"

        if (
            isinstance(self.engine, UnrealEngine)
            and os.path.commonpath(
                [os.path.realpath(file_path), self.engine.content_dir]
            )
            == self.engine.content_dir
        ):
            sub_path = (
                os.path.splitext(os.path.realpath(file_path))[0]
                .split(self.engine.content_dir)[1][1:]
                .replace(os.sep, "/")
            )
            return f"/Engine/{sub_path}"

        if self.plugins is not None:
            for plugin in self.plugins.values():
                unreal_path = plugin.unreal_path_from_file_path(file_path)
                if unreal_path is not None:
                    return unreal_path
        return None

    def validate(self) -> None:
        """Raise exceptions if this instance is misconfigured."""
        for entry_point in entry_points().get("crazyhusk.project.validators", []):
            entry_point.load()(self)



# Standard Library
from typing import Any, List, Optional, Tuple, Type

# Third Party
import pytest

# CrazyHusk
from crazyhusk import cli
from crazyhusk.__main__ import main


@pytest.mark.parametrize(
    "args,raises",
    [
        (None, (cli.CommandError, SystemExit, TypeError)),
        ([], (cli.CommandError, SystemExit)),
        (["test-command"], (cli.CommandError, SystemExit)),
        ([""], (cli.CommandError, SystemExit)),
        # TODO: monkeypatch pkg_resources behavior
    ],
)
def test_main(
    monkeypatch: Any,
    args: Optional[List[str]],
    raises: Optional[Tuple[Type[BaseException]]],
) -> None:
    monkeypatch.setattr("sys.argv", args)
    if raises is not None:
        with pytest.raises(raises):
            assert main()
    else:
        assert main()



"""Logging utilities for crazyhusk Unreal Engine object wrappers."""
# Standard Library
import calendar
import logging
import re
from datetime import datetime
from typing import Any

try:
    # Standard Library
    from typing import Literal  # type:ignore
except ImportError:
    # Third Party
    from typing_extensions import Literal  # type:ignore

# Regular expression for capturing data from UE4 logs
RE_UE4_LOG_LINE = re.compile(
    r"\[(?P<timestamp>[\d.:\-]+)\]\[.+\](?P<module>\w+): ?((?P<level>(Display)|(Info)|(Warning)|(Error)): ?)?(?P<message>.+)$"
)
RE_UBT_LOG_LINE = re.compile(
    r"^(?P<filename>.+?)\((?P<linenumber>\d+),?(?P<colnumber>\d+?)?\)\s?\:\s?(((?P<level>error|warning) \w+)|note)\:\s(?P<message>.+?)$"
)

UE4_LOG_MAP = {
    "Info": logging.INFO,
    "Display": logging.INFO,
    "Warning": logging.WARNING,
    "Error": logging.ERROR,
}

UBT_LOG_MAP = {"warning": logging.WARNING, "error": logging.ERROR}


class FilterEngineRun(logging.Filter):
    """Filter to enhance log records when using UnrealEngine.run()."""

    def __init__(self, executable: str, *args: str) -> None:
        """Initialize the filter base args."""
        logging.Filter.__init__(self)
        self.executable: str = executable
        self.cmd_args = args

    def filter(self, record: Any) -> Literal[True]:
        """Enhance a loggable record's attributes."""
        record.executable = self.executable
        record.cmd_args = self.cmd_args
        return True


class FilterUBTWarnings(logging.Filter):
    """Filter to enhance log records generated by UnrealBuildTool."""

    def filter(self, record: Any) -> Literal[True]:
        """Filter LogRecords emitted from UnrealBuildTool which are warnings/errors/notes."""
        captured = RE_UBT_LOG_LINE.match(record.msg)
        if captured is not None:
            record.levelno = UBT_LOG_MAP.get(captured.group("level"), logging.WARN)
            record.levelname = logging.getLevelName(record.levelno)
            record.filename = captured.group("filename")
            record.linenumber = captured.group("linenumber")
            record.colnumber = captured.group("colnumber")
            record.sub_msg = captured.group("message")
        return True


class FilterUE4Logs(logging.Filter):
    """Filter to enhance log records generated by UE4Editor/UE4Game."""

    def filter(self, record: Any) -> Literal[True]:
        """Filter LogRecords emitted from UE4Editor/UE4Game."""
        captured = RE_UE4_LOG_LINE.match(record.msg)
        if captured is not None:
            record.levelno = UE4_LOG_MAP.get(captured.group("level"), logging.INFO)
            record.levelname = logging.getLevelName(record.levelno)
            record.created = calendar.timegm(
                datetime.strptime(
                    captured.group("timestamp"), "%Y.%m.%d-%H.%M.%S:%f"
                ).timetuple()
            )
            record.module = captured.group("module")
            record.sub_msg = captured.group("message")
        return True



"""Expose crazyhusk functionality to the commandline."""
# Standard Library
import argparse
import inspect
import logging
import sys
from typing import Any, List

try:
    # Standard Library
    from importlib.metadata import entry_points  # type:ignore
except ImportError:
    # Third Party
    from importlib_metadata import entry_points  # type:ignore


class CommandError(Exception):
    """Custom exception representing errors encountered with CLI."""


def set_subcommand_arguments(
    parser: argparse.ArgumentParser, command: Any
) -> argparse.ArgumentParser:
    """Dynamically set argparse.Parser subcommand arguments by inspecting a callable function."""
    if not isinstance(parser, argparse.ArgumentParser):
        raise TypeError(
            f"Parser provided must be of type argparse.ArgumentParser. Got: {parser!r}"
        )

    if not inspect.isfunction(command):
        raise TypeError(
            f"Command provided must be a callable function. Got:{command!r}"
        )

    fullargs = inspect.getfullargspec(command)
    if fullargs.defaults is not None:
        firstdefault = len(fullargs.args) - len(fullargs.defaults)

    for i, arg in enumerate(fullargs.args):
        if fullargs.defaults and i >= firstdefault:
            default = fullargs.defaults[i - firstdefault]
            if default is None:
                parser.add_argument(dest=arg)
            elif type(default) is tuple or type(default) is list:
                parser.add_argument("--" + arg, default=default, nargs="+")
            elif type(default) is bool and default:
                parser.add_argument(
                    "--disable_" + arg, dest=arg, default=default, action="store_false"
                )
            elif type(default) is bool and not default:
                parser.add_argument(
                    "--enable_" + arg, dest=arg, default=default, action="store_true"
                )
            else:
                parser.add_argument("--" + arg, default=default, type=type(default))
        else:
            parser.add_argument(dest=arg)

    return parser


def parse_cli_args(args: List[str]) -> argparse.Namespace:
    """Parse crazyhusk CLI arguments."""
    if args is None:
        raise CommandError("None is not parsable arguments.")
    elif len(args) == 0:
        raise CommandError("Must provide at least one argument.")

    parser = argparse.ArgumentParser()
    commands_parser = parser.add_subparsers(
        title="subcommands", description="valid subcommands", help="subcommand help"
    )

    for entry_point in entry_points().get("crazyhusk.commands", []):
        command = entry_point.load()
        if inspect.isfunction(command):
            cmd_parser = commands_parser.add_parser(entry_point.name)
            cmd_parser.set_defaults(command=command)
            set_subcommand_arguments(cmd_parser, command)

    return parser.parse_args(args)


def run(args: List[str] = sys.argv[1:]) -> None:
    """Run the crazyhusk CLI entrypoint."""
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    command_args = parse_cli_args(args)
    if "command" in command_args:
        command_args.command(
            **{k: v for k, v in command_args.__dict__.items() if k != "command"}
        )



"""Utilities for working with report formats generated by Unreal Engine."""

# Standard Library
import datetime
import json
import os
from typing import Any, Dict, List, Optional
from xml.dom import minidom
from xml.etree import ElementTree
from xml.etree.ElementTree import Element


def report_timestamp_to_iso8601_timestamp(timestamp: str) -> str:
    return datetime.datetime.strptime(timestamp, "%Y.%m.%d-%H.%M.%S").isoformat()


def report_entry_to_entry_xml(entry: Dict[str, Any]) -> Element:
    failure = Element("failure")
    failure.set("type", entry.get("event", {}).get("type", ""))
    timestamp = entry.get("timestamp")
    if timestamp is not None:
        failure.set("timestamp", report_timestamp_to_iso8601_timestamp(timestamp))
    failure.set(
        "message",
        f'{entry.get("filename","")}:{entry.get("lineNumber","")} {entry.get("event",{}).get("message","")}',
    )
    return failure


def report_test_to_testcase_xml(test: Dict[str, Any]) -> Element:
    test_case = Element("testcase")
    test_case.set("name", test.get("testDisplayName", ""))
    test_case.set("classname", test.get("fullTestPath", ""))
    test_case.set("status", test.get("state", ""))
    for entry in test.get("entries", []):
        test_case.append(report_entry_to_entry_xml(entry))
    return test_case


def report_object_to_testsuite_xml(report: Dict[str, Any]) -> Element:
    test_suite = Element("testsuite")
    test_suite.set("tests", str(len(report.get("tests", []))))
    test_suite.set("failures", str(report.get("failed", 0)))
    test_suite.set("skipped", str(report.get("notRun", 0)))
    test_suite.set("time", str(report.get("totalDuration", 0.0)))
    timestamp = report.get("reportCreatedOn")
    if timestamp is not None:
        test_suite.set(
            "timestamp",
            report_timestamp_to_iso8601_timestamp(timestamp),
        )

    branch, changelist, platform = report.get("clientDescriptor", " -  - ").split(" - ")
    properties = Element("properties")

    branch_property = Element("property")
    branch_property.set("name", "branch")
    branch_property.set("value", branch)
    properties.append(branch_property)

    changelist_property = Element("property")
    changelist_property.set("name", "changelist")
    changelist_property.set("value", changelist)
    properties.append(changelist_property)

    platform_property = Element("property")
    platform_property.set("name", "platform")
    platform_property.set("value", platform)
    properties.append(platform_property)

    test_suite.append(properties)

    for test in report.get("tests", []):
        test_suite.append(report_test_to_testcase_xml(test))

    return test_suite


def json_report_to_dict(report_file: str) -> Dict[str, Any]:
    if not os.path.isfile(report_file):
        raise ValueError(f"JSON report not found: {report_file}")

    if not os.path.splitext(report_file)[-1] == ".json":
        raise ValueError(f"Report file is not JSON: {report_file}")

    with open(report_file, "r", encoding="utf-8-sig") as json_report:
        report_dct = json.load(json_report)
        if not isinstance(report_dct, dict):
            raise ValueError(f"JSON report returns non-object: {report_file}")
        return report_dct


def write_junit_xml_report(report_file: str, test_suites: Element) -> None:
    if not os.path.splitext(report_file)[-1] == ".xml":
        raise ValueError(f"Report file is not XML: {report_file}")

    report_dir = os.path.dirname(report_file)
    if not os.path.isdir(report_dir):
        os.makedirs(report_dir, exist_ok=True)

    with open(report_file, "w", encoding="utf-8") as xml_report:
        xml_report.write(
            minidom.parseString(ElementTree.tostring(test_suites, "utf-8")).toprettyxml(
                indent=" " * 4
            )
        )


def json_reports_to_junit_xml(
    junit_file: Optional[str] = None, json_reports: List[str] = []
) -> None:
    """Convert a JSON report from Unreal automation to jUnit XML format."""
    test_suites = Element("testsuites")
    test_suites.set("name", "Unreal Automation Tests")

    total_tests = 0
    total_failures = 0
    total_errors = 0
    total_time = 0.0
    for report in json_reports:
        test_suite = report_object_to_testsuite_xml(json_report_to_dict(report))
        test_suite.set("name", os.path.splitext(os.path.basename(report))[0])
        total_tests += int(test_suite.get("tests", 0))
        total_failures += int(test_suite.get("failures", 0))
        total_errors += int(test_suite.get("errors", 0))
        total_time += float(test_suite.get("time", 0.0))
        test_suites.append(test_suite)

    test_suites.set("tests", str(total_tests))
    test_suites.set("failures", str(total_failures))
    test_suites.set("errors", str(total_errors))
    test_suites.set("time", str(total_time))
    if junit_file is not None:
        write_junit_xml_report(junit_file, test_suites)



# Standard Library
import argparse
import inspect
from typing import Any, List, Optional, Tuple, Type

# Third Party
import pytest

# CrazyHusk
from crazyhusk import cli


def test_set_subcommand_arguments_types() -> None:
    with pytest.raises(TypeError):
        cli.set_subcommand_arguments(None, None)

    with pytest.raises(TypeError):
        cli.set_subcommand_arguments(argparse.ArgumentParser(), None)

    def subcommand_noargs() -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(argparse.ArgumentParser(), subcommand_noargs),
        argparse.ArgumentParser,
    )

    def subcommand_nodefaults(arg: Any) -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(argparse.ArgumentParser(), subcommand_nodefaults),
        argparse.ArgumentParser,
    )

    def subcommand_default_none(arg: Optional[Any] = None) -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(
            argparse.ArgumentParser(), subcommand_default_none
        ),
        argparse.ArgumentParser,
    )

    def subcommand_default_tuple(arg: Any = ()) -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(
            argparse.ArgumentParser(), subcommand_default_tuple
        ),
        argparse.ArgumentParser,
    )

    def subcommand_default_list(arg: Any = []) -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(
            argparse.ArgumentParser(), subcommand_default_list
        ),
        argparse.ArgumentParser,
    )

    def subcommand_default_true(arg: Any = True) -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(
            argparse.ArgumentParser(), subcommand_default_true
        ),
        argparse.ArgumentParser,
    )

    def subcommand_default_false(arg: Any = False) -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(
            argparse.ArgumentParser(), subcommand_default_false
        ),
        argparse.ArgumentParser,
    )

    def subcommand_default_string(arg: Any = "") -> None:
        pass

    assert isinstance(
        cli.set_subcommand_arguments(
            argparse.ArgumentParser(), subcommand_default_string
        ),
        argparse.ArgumentParser,
    )


@pytest.mark.parametrize(
    "args,raises",
    [
        (None, cli.CommandError),
        ([], cli.CommandError),
        (["test-command"], SystemExit),
        ([""], SystemExit),
        (["test"], None),
    ],
)
def test_parse_cli_args(
    args: Optional[List[str]],
    raises: Optional[Type[BaseException]],
    monkeypatch: Any,
    test_entry_point: Any,
) -> None:
    monkeypatch.setattr(
        cli, "entry_points", lambda: {"crazyhusk.commands": [test_entry_point]}
    )
    if raises is not None:
        with pytest.raises(raises):
            assert cli.parse_cli_args(args)
    else:
        parsed = cli.parse_cli_args(args)
        assert isinstance(parsed, argparse.Namespace)
        assert "command" in parsed
        assert inspect.isfunction(parsed.command)


def test_parse_cli_args_entry_points(
    monkeypatch: Any, test_entry_point: Any, null_entry_point: Any
) -> None:
    monkeypatch.setattr(
        cli, "entry_points", lambda: {"crazyhusk.commands": [test_entry_point]}
    )
    args = cli.parse_cli_args(["test"])
    assert isinstance(args, argparse.Namespace)
    assert "command" in args

    monkeypatch.setattr(
        cli, "entry_points", lambda: {"crazyhusk.commands": [null_entry_point]}
    )
    with pytest.raises(SystemExit):
        assert cli.parse_cli_args(["test"]) is None


@pytest.mark.parametrize(
    "args,raises",
    [
        (None, (cli.CommandError, SystemExit)),
        ([], (cli.CommandError, SystemExit)),
        (["test-command"], (cli.CommandError, SystemExit)),
        ([""], (cli.CommandError, SystemExit)),
        (["test"], None),
    ],
)
def test_cli_run(
    args: Optional[List[str]],
    raises: Optional[Tuple[Type[BaseException]]],
    monkeypatch: Any,
    test_entry_point: Any,
) -> None:
    monkeypatch.setattr(
        cli, "entry_points", lambda: {"crazyhusk.commands": [test_entry_point]}
    )
    if raises is not None:
        with pytest.raises(raises):
            assert cli.run(args) is None
    else:
        assert cli.run(args) is None


def test_cli_run_entry_points(monkeypatch: Any, test_entry_point: Any) -> None:
    monkeypatch.setattr(
        cli, "entry_points", lambda: {"crazyhusk.commands": [test_entry_point]}
    )
    assert cli.run(["test"]) is None



# Standard Library
from typing import Any, Optional, Set, Type

# Third Party
import pytest

# CrazyHusk
from crazyhusk import code


@pytest.mark.parametrize(
    "code_template_fixture,name,template_string,tokens",
    [
        ("null_code_template", None, "", set()),
        ("empty_code_template", "", "", set()),
        ("basic_code_template", "Basic", "%TEST_TOKEN%", {"TEST_TOKEN"}),
        (
            "multiline_basic_code_template",
            "MultilineBasic",
            "//%TEST_TOKEN%\n//%TEST_TOKEN%",
            {"TEST_TOKEN"},
        ),
        (
            "multiline_multitoken_code_template",
            "MultilineMultitoken",
            r"//%TEST_TOKEN_1%\n//%TEST_TOKEN_2%",
            set(),
        ),
    ],
)
def test_codetemplate_init(
    code_template_fixture: str,
    name: Optional[str],
    template_string: str,
    tokens: Set[str],
    request: Any,
) -> None:
    code_template = request.getfixturevalue(code_template_fixture)
    assert code_template.name == name
    assert code_template.template_string == template_string
    assert repr(code_template) == f"<CodeTemplate {code_template.name}>"
    assert type(code_template.tokens) is set
    assert code_template.tokens == tokens


@pytest.mark.parametrize(
    "code_template_fixture,tokens,raises,expected",
    [
        ("null_code_template", {}, None, ""),
        ("empty_code_template", {}, None, ""),
        ("basic_code_template", {}, code.CodeTemplateError, ""),
        ("basic_code_template", {"TEST_TOKEN": "test"}, None, "test"),
        (
            "multiline_basic_code_template",
            {"TEST_TOKEN": "test"},
            None,
            "//test\n//test",
        ),
        (
            "multiline_multitoken_code_template",
            {"TEST_TOKEN_1": "test1", "TEST_TOKEN_2": "test2"},
            None,
            r"//test1\n//test2",
        ),
    ],
)
def test_codetemplate_make_instance(
    code_template_fixture: str,
    tokens: Set[str],
    raises: Optional[Type[BaseException]],
    expected: str,
    request: Any,
) -> None:
    code_template = request.getfixturevalue(code_template_fixture)
    if raises is not None:
        with pytest.raises(raises):
            code_template.make_instance(**tokens)
    else:
        assert code_template.make_instance(**tokens) == expected



# Standard Library
import os
import types
from typing import Any, Dict, Optional, Type

# Third Party
import pytest

# CrazyHusk
from crazyhusk import config, engine
from crazyhusk.code import CodeTemplate
from crazyhusk.plugin import UnrealPlugin


@pytest.mark.parametrize(
    "version_fixture,major,minor,patch,changelist,branch",
    [
        ("version_empty", 4, 0, 0, 0, ""),
        ("version_26_0", 4, 26, 0, 0, ""),
        ("version_26_1", 4, 26, 1, 0, ""),
        ("version_26_1_123456", 4, 26, 1, 123456, ""),
        ("version_26_1_123456_branch", 4, 26, 1, 123456, "++UE4+Release-4.26"),
    ],
)
def test_unreal_version_init(
    version_fixture: str,
    major: int,
    minor: int,
    patch: int,
    changelist: int,
    branch: str,
    request: Any,
) -> None:
    version = request.getfixturevalue(version_fixture)
    assert version.major == major
    assert version.minor == minor
    assert version.patch == patch
    assert version.changelist == changelist
    assert version.branch == branch


@pytest.mark.parametrize(
    "version_fixture,version_string",
    [
        ("version_empty", "4.0"),
        ("version_26_0", "4.26"),
        ("version_26_1", "4.26.1"),
        ("version_26_1_123456", "4.26.1-123456"),
        ("version_26_1_123456_branch", "4.26.1-123456+++UE4+Release-4.26"),
    ],
)
def test_unreal_version_string(
    version_fixture: str, version_string: str, request: Any
) -> None:
    version = request.getfixturevalue(version_fixture)
    assert str(version) == version_string


@pytest.mark.parametrize(
    "version_fixture,version_repr",
    [
        ("version_empty", "<UnrealVersion 4.0>"),
        ("version_26_0", "<UnrealVersion 4.26>"),
        ("version_26_1", "<UnrealVersion 4.26.1>"),
        ("version_26_1_123456", "<UnrealVersion 4.26.1-123456>"),
        (
            "version_26_1_123456_branch",
            "<UnrealVersion 4.26.1-123456+++UE4+Release-4.26>",
        ),
    ],
)
def test_unreal_version_repr(
    version_fixture: str, version_repr: str, request: Any
) -> None:
    version = request.getfixturevalue(version_fixture)
    assert repr(version) == version_repr


@pytest.mark.parametrize(
    "first_version_fixture,second_version_fixture",
    [
        ("version_empty", "version_26_0"),
        ("version_empty", "version_26_1"),
        ("version_empty", "version_26_1_123456"),
        ("version_empty", "version_26_1_123456_branch"),
        ("version_empty", "version_5_0"),
        ("version_26_0", "version_26_1"),
        ("version_26_0", "version_26_1_123456"),
        ("version_26_0", "version_26_1_123456_branch"),
        ("version_26_0", "version_5_0"),
        ("version_26_1", "version_26_1_123456"),
        ("version_26_1", "version_26_1_123456_branch"),
        ("version_26_1", "version_5_0"),
        ("version_26_1_123456", "version_26_1_234567"),
        ("version_26_1_123456", "version_5_0"),
        ("version_26_1_234567", "version_5_0"),
    ],
)
def test_unreal_version_lt(
    first_version_fixture: str, second_version_fixture: str, request: Any
) -> None:
    first_version = request.getfixturevalue(first_version_fixture)
    second_version = request.getfixturevalue(second_version_fixture)
    assert first_version < second_version


def test_unreal_version_lt_null(version_empty: engine.UnrealVersion) -> None:
    with pytest.raises(TypeError):
        assert version_empty < None


def test_unreal_version_lt_not(
    version_empty: engine.UnrealVersion, version_5_0: engine.UnrealVersion
) -> None:
    assert not version_5_0 < version_empty


@pytest.mark.parametrize(
    "first_version_fixture,second_version_fixture",
    [
        ("version_empty", "version_empty"),
        ("version_26_0", "version_26_0"),
        ("version_26_1", "version_26_1"),
        ("version_26_1_123456", "version_26_1_123456"),
        ("version_26_1_234567", "version_26_1_234567"),
        ("version_5_0", "version_5_0"),
    ],
)
def test_unreal_version_eq(
    first_version_fixture: str, second_version_fixture: str, request: Any
) -> None:
    first_version = request.getfixturevalue(first_version_fixture)
    second_version = request.getfixturevalue(second_version_fixture)
    assert first_version == second_version


def test_unreal_version_eq_null(version_empty: engine.UnrealVersion) -> None:
    assert not version_empty == None


@pytest.mark.parametrize(
    "dct,version_string",
    [
        ({}, "4.0"),
        ({"MajorVersion": 5}, "5.0"),
        ({"MinorVersion": 26}, "4.26"),
        ({"MinorVersion": 26, "PatchVersion": 1}, "4.26.1"),
    ],
)
def test_unreal_version_to_object(dct: Dict[str, int], version_string: str) -> None:
    assert str(engine.UnrealVersion.to_object(dct)) == version_string


# Tests for UnrealEngine object
@pytest.mark.parametrize(
    "base_dir,raises,expected",
    [
        (None, engine.UnrealEngineError, None),
        ("", engine.UnrealEngineError, None),
        (".", None, os.path.realpath(".")),
    ],
)
def test_unreal_engine_init(
    base_dir: Optional[str],
    raises: Optional[Type[BaseException]],
    expected: Optional[str],
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert engine.UnrealEngine(base_dir)
    else:
        assert engine.UnrealEngine(base_dir).base_dir == expected


def test_unreal_engine_repr(engine_empty: engine.UnrealEngine) -> None:
    assert (
        repr(engine_empty)
        == f"<UnrealEngine None Build None at {engine_empty.base_dir}>"
    )


def test_unreal_engine_lt_types(
    engine_empty: engine.UnrealEngine, engine_empty_version_empty: engine.UnrealEngine
) -> None:
    with pytest.raises(TypeError):
        assert engine_empty < None
        assert engine_empty < engine_empty_version_empty


def test_unreal_engine_lt(
    engine_empty_version_empty: engine.UnrealEngine,
    engine_empty_version_egl_4_26_2: engine.UnrealEngine,
) -> None:
    assert engine_empty_version_empty < engine_empty_version_egl_4_26_2


def test_unreal_engine_context(engine_empty: engine.UnrealEngine) -> None:
    with engine_empty:
        assert engine_empty


def test_unreal_engine_dir_properties(
    engine_empty_version_egl_4_26_2: engine.UnrealEngine,
) -> None:
    assert engine_empty_version_egl_4_26_2.engine_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "Engine"
    )
    assert engine_empty_version_egl_4_26_2.feature_packs_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "FeaturePacks"
    )
    assert engine_empty_version_egl_4_26_2.samples_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "Samples"
    )
    assert engine_empty_version_egl_4_26_2.templates_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "Templates"
    )
    assert engine_empty_version_egl_4_26_2.build_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "Engine", "Build"
    )
    assert engine_empty_version_egl_4_26_2.config_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "Engine", "Config"
    )
    assert engine_empty_version_egl_4_26_2.content_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "Engine", "Content"
    )
    assert engine_empty_version_egl_4_26_2.plugins_dir == os.path.join(
        engine_empty_version_egl_4_26_2.base_dir, "Engine", "Plugins"
    )


@pytest.mark.parametrize(
    "unreal_engine_fixture,expected",
    [
        ("engine_empty", None),
        ("engine_empty_version_egl_4_26_2", "Installed"),
        ("engine_empty_version_empty", "Source"),
    ],
)
def test_unreal_engine_build_type(
    unreal_engine_fixture: str, expected: Optional[str], request: Any
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    assert unreal_engine.build_type == expected


@pytest.mark.parametrize(
    "unreal_engine_fixture,executable,raises",
    [
        ("engine_local", None, TypeError),
        ("engine_local", ".", None),
        ("engine_local", "/bin/scary_dir/important_exe", engine.UnrealExecutionError),
        ("engine_local", os.path.realpath(".."), engine.UnrealExecutionError),
    ],
)
def test_unreal_engine_exe_common_path(
    unreal_engine_fixture: str,
    executable: Optional[str],
    raises: Optional[Type[BaseException]],
    request: Any,
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert engine.UnrealEngine.engine_exe_common_path(unreal_engine, executable)
    else:
        assert (
            engine.UnrealEngine.engine_exe_common_path(unreal_engine, executable)
            is None
        )


def test_unreal_engine_exe_common_path_types() -> None:
    with pytest.raises(TypeError):
        assert engine.UnrealEngine.engine_exe_common_path(None, None)


@pytest.mark.parametrize(
    "unreal_engine_fixture,executable,raises",
    [
        ("engine_empty", "", engine.UnrealExecutionError),
        ("engine_empty_version_egl_4_26_2", "", engine.UnrealExecutionError),
        ("engine_empty_version_empty", "", engine.UnrealExecutionError),
        ("engine_local", None, TypeError),
        ("engine_local", ".", engine.UnrealExecutionError),
        ("engine_local", "/bin/scary_dir/important_exe", engine.UnrealExecutionError),
        ("engine_local", r"C:\Windows\scaryfile.exe", engine.UnrealExecutionError),
        ("engine_local", os.path.realpath(".."), engine.UnrealExecutionError),
    ],
)
def test_unreal_engine_exe_exists(
    unreal_engine_fixture: str,
    executable: Optional[str],
    raises: Optional[Type[BaseException]],
    request: Any,
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert engine.UnrealEngine.engine_exe_exists(unreal_engine, executable)
    else:
        assert engine.UnrealEngine.engine_exe_exists(unreal_engine, executable) is None


def test_unreal_engine_exe_exists_types() -> None:
    with pytest.raises(TypeError):
        assert engine.UnrealEngine.engine_exe_exists(None, None)


@pytest.mark.parametrize(
    "unreal_engine_fixture,raises",
    [
        ("engine_empty", engine.UnrealEngineError),
        ("engine_empty_version_egl_4_26_2", None),
        ("engine_empty_version_empty", None),
        ("engine_local", engine.UnrealEngineError),
    ],
)
def test_unreal_engine_dir_exists(
    unreal_engine_fixture: str, raises: Optional[Type[BaseException]], request: Any
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert engine.UnrealEngine.engine_dir_exists(unreal_engine)
    else:
        assert engine.UnrealEngine.engine_dir_exists(unreal_engine) is None


def test_unreal_engine_dir_exist_types() -> None:
    with pytest.raises(TypeError):
        assert engine.UnrealEngine.engine_dir_exists(None)


@pytest.mark.parametrize(
    "unreal_engine_fixture,expected",
    [
        ("engine_empty", False),
        ("engine_empty_version_egl_4_26_2", True),
        ("engine_empty_version_empty", False),
        ("engine_local", False),
    ],
)
def test_unreal_engine_is_installed_build(
    unreal_engine_fixture: str, expected: bool, request: Any
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    assert unreal_engine.is_installed_build() is expected


@pytest.mark.parametrize(
    "unreal_engine_fixture,expected",
    [
        ("engine_empty", False),
        ("engine_empty_version_egl_4_26_2", False),
        ("engine_empty_version_empty", True),
        ("engine_local", False),
    ],
)
def test_unreal_engine_is_source_build(
    unreal_engine_fixture: str, expected: bool, request: Any
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    assert unreal_engine.is_source_build() is expected


@pytest.mark.parametrize(
    "unreal_path,raises,expected",
    [
        ("", engine.UnrealEngineError, None),
        ("/", engine.UnrealEngineError, None),
        ("/Game", engine.UnrealEngineError, None),
        ("/Game/whatever", engine.UnrealEngineError, None),
        ("/Engine", engine.UnrealEngineError, None),
        (
            "/Engine/whatever",
            None,
            os.path.realpath(os.path.join(".", "Engine", "Content", "whatever.uasset")),
        ),
        ("/Invalid", engine.UnrealEngineError, None),
        ("/Invalid/whatever", engine.UnrealEngineError, None),
        ("/test", engine.UnrealEngineError, None),
        ("/test/whatever", engine.UnrealEngineError, None),
    ],
)
def test_unreal_path_to_file_path(
    unreal_path: str,
    raises: Optional[Type[BaseException]],
    expected: Optional[str],
    engine_local: engine.UnrealEngine,
) -> None:
    if raises:
        with pytest.raises(raises):
            assert engine_local.unreal_path_to_file_path(unreal_path) == expected
    else:
        assert engine_local.unreal_path_to_file_path(unreal_path) == expected


@pytest.mark.parametrize(
    "file_path,raises,expected",
    [
        ("", engine.UnrealEngineError, None),
        ("/", engine.UnrealEngineError, None),
        (".", engine.UnrealEngineError, None),
        (os.path.realpath("./Engine"), engine.UnrealEngineError, None),
        (os.path.realpath("./Engine/Content"), None, "/Engine/"),
        (
            os.path.realpath(os.path.join(".", "Engine", "Content", "whatever.uasset")),
            None,
            "/Engine/whatever",
        ),
    ],
)
def test_unreal_path_from_file_path(
    file_path: str,
    raises: Optional[Type[BaseException]],
    expected: Optional[str],
    engine_local: engine.UnrealEngine,
) -> None:
    if raises:
        with pytest.raises(raises):
            assert engine_local.unreal_path_from_file_path(file_path) == expected
    else:
        assert engine_local.unreal_path_from_file_path(file_path) == expected


@pytest.mark.parametrize(
    "unreal_engine_fixture,expected",
    [
        ("engine_empty", {}),
        ("engine_empty_version_egl_4_26_2", {}),
        ("engine_empty_version_empty", {}),
        ("engine_local", {}),
    ],
)
def test_unreal_engine_code_templates(
    unreal_engine_fixture: str, expected: Dict[str, CodeTemplate], request: Any
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    assert unreal_engine.code_templates == expected


@pytest.mark.parametrize(
    "unreal_engine_fixture,expected",
    [
        ("engine_empty", {}),
        ("engine_empty_version_egl_4_26_2", {}),
        ("engine_empty_version_empty", {}),
        ("engine_local", {}),
    ],
)
def test_unreal_engine_plugins(
    unreal_engine_fixture: str, expected: Dict[str, UnrealPlugin], request: Any
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    assert unreal_engine.plugins == expected


@pytest.mark.parametrize(
    "association,expected",
    [
        ("randomstring", type(None)),
        # TODO: monkeypatch pkg_resources behavior
    ],
)
def test_unreal_engine_find_engine(association: str, expected: Type[Any]) -> None:
    assert type(engine.UnrealEngine.find_engine(association)) == expected


def test_unreal_engine_list_all_engines() -> None:
    assert isinstance(engine.UnrealEngine.list_all_engines(), types.GeneratorType)

    # TODO: monkeypatch pkg_resources behavior
    # assert list(engine.UnrealEngine.list_all_engines())


@pytest.mark.parametrize(
    "unreal_engine_fixture,config_category,platform,expected_count",
    [
        ("engine_empty_version_egl_4_26_2", None, None, 1),
        ("engine_empty_version_egl_4_26_2", "", None, 1),
        ("engine_empty_version_egl_4_26_2", "test", None, 1),
        ("engine_empty_version_egl_4_26_2", "Engine", None, 2),
        ("engine_empty_version_egl_4_26_2", "Engine", "Windows", 4),
    ],
)
def test_unreal_engine_config_files(
    unreal_engine_fixture: str,
    config_category: Optional[str],
    platform: Optional[str],
    expected_count: int,
    request: Any,
) -> None:
    unreal_engine = request.getfixturevalue(unreal_engine_fixture)
    assert (
        len(list(unreal_engine.config_files(config_category, platform)))
        == expected_count
    )


def test_unreal_engine_config(
    engine_empty_version_egl_4_26_2: engine.UnrealEngine,
) -> None:
    assert isinstance(
        engine_empty_version_egl_4_26_2.config(), config.UnrealConfigParser
    )
    assert isinstance(
        engine_empty_version_egl_4_26_2.config("Engine"), config.UnrealConfigParser
    )
    assert isinstance(
        engine_empty_version_egl_4_26_2.config("Engine", "Windows"),
        config.UnrealConfigParser,
    )



"""
Invokes crazyhusk CLI when the crazyhusk module is run as a script.

Example: python -m crazyhusk list_engines
"""


def main() -> None:
    # CrazyHusk
    from crazyhusk import cli

    cli.run()


if __name__ == "__main__":
    main()



# Standard Library
import os
import types
from typing import Any, Dict, Optional, Type

# Third Party
import pytest

# CrazyHusk
from crazyhusk import code, config, engine, module, plugin, project


def test_project_descriptor_init(
    null_association_project_descriptor: project.ProjectDescriptor,
) -> None:
    assert null_association_project_descriptor.engine_association is None
    assert null_association_project_descriptor.category == ""
    assert null_association_project_descriptor.description == ""
    assert not null_association_project_descriptor.disable_engine_plugins_by_default
    assert not null_association_project_descriptor.is_enterprise_project
    assert null_association_project_descriptor.epic_sample_name_hash is None
    assert null_association_project_descriptor.post_build_steps is None
    assert null_association_project_descriptor.pre_build_steps is None
    assert isinstance(null_association_project_descriptor.target_platforms, list)
    assert len(null_association_project_descriptor.target_platforms) == 0
    assert isinstance(null_association_project_descriptor.modules, types.GeneratorType)
    assert isinstance(null_association_project_descriptor.plugins, types.GeneratorType)


def test_project_descriptor_repr(
    null_association_project_descriptor: project.ProjectDescriptor,
) -> None:
    assert repr(null_association_project_descriptor) == "<ProjectDescriptor >"


@pytest.mark.parametrize(
    "project_descriptor_fixture,expected_type",
    [
        ("null_association_project_descriptor", dict),
        ("empty_association_project_descriptor", project.ProjectDescriptor),
        ("basic_project_descriptor", project.ProjectDescriptor),
    ],
)
def test_project_descriptor_to_object(
    project_descriptor_fixture: str, expected_type: Type[Any], request: Any
) -> None:
    project_descriptor = request.getfixturevalue(project_descriptor_fixture)
    dct = project_descriptor.to_dict()
    assert isinstance(project.ProjectDescriptor.to_object(dct), expected_type)


def test_project_descriptor_to_object_module_dict(
    basic_project_descriptor_withmodule_dict: Dict[str, Any],
) -> None:
    _project = project.ProjectDescriptor.to_object(
        basic_project_descriptor_withmodule_dict
    )
    assert isinstance(_project, project.ProjectDescriptor)
    for _module in _project.modules:
        assert isinstance(_module, module.ModuleDescriptor)


@pytest.mark.parametrize(
    "project_descriptor_fixture,expected_type",
    [
        ("null_association_project_descriptor", dict),
        ("empty_association_project_descriptor", project.ProjectDescriptor),
        ("basic_project_descriptor", project.ProjectDescriptor),
    ],
)
def test_project_descriptor_modules(
    project_descriptor_fixture: str, expected_type: Type[Any], request: Any
) -> None:
    project_descriptor = request.getfixturevalue(project_descriptor_fixture)
    for _module in project_descriptor.modules:
        assert isinstance(_module, expected_type)


def test_project_descriptor_add_module(
    basic_project_descriptor: project.ProjectDescriptor,
    default_valid_module_descriptor: module.ModuleDescriptor,
) -> None:
    with pytest.raises(NotImplementedError):
        assert basic_project_descriptor.add_module({}) is None
    assert basic_project_descriptor.add_module(default_valid_module_descriptor) is None


# UnrealProject tests


@pytest.mark.parametrize(
    "unreal_project_fixture,raises",
    [
        ("null_filename_unreal_project", TypeError),
        ("empty_filename_unreal_project", None),
        ("invalid_filename_unreal_project", None),
        ("empty_file_unreal_project", None),
    ],
)
def test_unreal_project_init_args(
    unreal_project_fixture: str, raises: Optional[Type[BaseException]], request: Any
) -> None:
    if raises is not None:
        with pytest.raises(raises):
            assert request.getfixturevalue(unreal_project_fixture) is not None
    else:
        assert request.getfixturevalue(unreal_project_fixture) is not None


def test_unreal_project_init(empty_file_unreal_project: project.UnrealProject) -> None:
    assert empty_file_unreal_project.project_file == "MyProject.uproject"
    assert empty_file_unreal_project.name == "MyProject"


def test_unreal_project_repr(empty_file_unreal_project: project.UnrealProject) -> None:
    assert (
        repr(empty_file_unreal_project)
        == "<UnrealProject MyProject at MyProject.uproject>"
    )


def test_unreal_project_properties(
    empty_file_unreal_project: project.UnrealProject,
) -> None:
    assert empty_file_unreal_project.project_dir == ""
    assert empty_file_unreal_project.config_dir == "Config"
    assert empty_file_unreal_project.content_dir == "Content"
    assert empty_file_unreal_project.plugins_dir == "Plugins"
    assert empty_file_unreal_project.saved_dir == "Saved"


@pytest.mark.parametrize(
    "unreal_project_fixture,raises",
    [
        ("empty_filename_unreal_project", project.UnrealProjectError),
        ("invalid_filename_unreal_project", project.UnrealProjectError),
        ("empty_file_content_unreal_project", None),
    ],
)
def test_unreal_project_file_exists(
    unreal_project_fixture: str, raises: Optional[Type[BaseException]], request: Any
) -> None:
    unreal_project = request.getfixturevalue(unreal_project_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert project.UnrealProject.project_file_exists(unreal_project) is None
    else:
        assert project.UnrealProject.project_file_exists(unreal_project) is None


def test_unreal_project_file_exists_types() -> None:
    with pytest.raises(TypeError):
        assert project.UnrealProject.project_file_exists(None) is None


@pytest.mark.parametrize(
    "unreal_project_fixture,raises",
    [
        ("empty_filename_unreal_project", project.UnrealProjectError),
        ("invalid_filename_unreal_project", project.UnrealProjectError),
        ("empty_file_content_unreal_project", None),
    ],
)
def test_unreal_project_valid_project_file_extension(
    unreal_project_fixture: str, raises: Optional[Type[BaseException]], request: Any
) -> None:
    unreal_project = request.getfixturevalue(unreal_project_fixture)
    if raises is not None:
        with pytest.raises(raises):
            assert (
                project.UnrealProject.valid_project_file_extension(unreal_project)
                is None
            )
    else:
        assert (
            project.UnrealProject.valid_project_file_extension(unreal_project) is None
        )


def test_unreal_project_valid_project_file_extension_types() -> None:
    with pytest.raises(TypeError):
        assert project.UnrealProject.valid_project_file_extension(None) is None


@pytest.mark.parametrize(
    "unreal_project_fixture,expected_type",
    [
        ("basic_unreal_project", module.ModuleDescriptor),
    ],
)
def test_unreal_project_modules(
    unreal_project_fixture: str, expected_type: Type[Any], request: Any
) -> None:
    unreal_project = request.getfixturevalue(unreal_project_fixture)
    for _name, _module in unreal_project.modules.items():
        assert isinstance(_name, str)
        assert isinstance(_module, expected_type)


def test_unreal_project_descriptor(
    basic_unreal_project: project.UnrealProject, monkeypatch: Any
) -> None:
    monkeypatch.setattr(project, "entry_points", lambda: {})
    assert isinstance(basic_unreal_project.descriptor, project.ProjectDescriptor)


@pytest.mark.parametrize(
    "unreal_path,raises,expected",
    [
        ("", project.UnrealProjectError, None),
        ("/", project.UnrealProjectError, None),
        ("/Game", project.UnrealProjectError, None),
        (
            "/Game/whatever",
            None,
            os.path.realpath(os.path.join(".", "Content", "whatever.uasset")),
        ),
        ("/Engine", project.UnrealProjectError, None),
        (
            "/Engine/whatever",
            None,
            os.path.realpath(
                os.path.join("..", "Engine", "Content", "whatever.uasset")
            ),
        ),
        ("/Invalid", project.UnrealProjectError, None),
        ("/Invalid/whatever", project.UnrealProjectError, None),
        ("/test", project.UnrealProjectError, None),
        ("/test/whatever", project.UnrealProjectError, None),
    ],
)
def test_unreal_path_to_file_path(
    unreal_path: str,
    raises: Optional[Type[BaseException]],
    expected: Optional[str],
    basic_unreal_project_realpath: project.UnrealProject,
) -> None:
    if raises:
        with pytest.raises(raises):
            assert (
                basic_unreal_project_realpath.unreal_path_to_file_path(unreal_path)
                == expected
            )
    else:
        assert (
            basic_unreal_project_realpath.unreal_path_to_file_path(unreal_path)
            == expected
        )


@pytest.mark.parametrize(
    "file_path,raises,expected",
    [
        ("", None, None),
        ("/", None, None),
        (".", None, None),
        (os.path.realpath("."), None, None),
        (os.path.realpath("./Content"), None, "/Game/"),
        (
            os.path.realpath(os.path.join(".", "Content", "whatever.uasset")),
            None,
            "/Game/whatever",
        ),
        (os.path.realpath("../Engine/Content"), None, "/Engine/"),
        (
            os.path.realpath(
                os.path.join("..", "Engine", "Content", "whatever.uasset")
            ),
            None,
            "/Engine/whatever",
        ),
    ],
)
def test_unreal_path_from_file_path(
    file_path: str,
    raises: Optional[Type[BaseException]],
    expected: Optional[str],
    basic_unreal_project_realpath: project.UnrealProject,
) -> None:
    if raises:
        with pytest.raises(raises):
            assert (
                basic_unreal_project_realpath.unreal_path_from_file_path(file_path)
                == expected
            )
    else:
        assert (
            basic_unreal_project_realpath.unreal_path_from_file_path(file_path)
            == expected
        )


def test_unreal_project_plugins(
    basic_unreal_project_realpath: project.UnrealProject,
) -> None:
    for _name, _plugin in basic_unreal_project_realpath.plugins.items():
        assert isinstance(_name, str)
        assert isinstance(_plugin, plugin.UnrealPlugin)


def test_unreal_project_code_templates(
    basic_unreal_project_realpath: project.UnrealProject, monkeypatch: Any
) -> None:
    monkeypatch.setattr(project, "entry_points", lambda: {})
    for _name, _plugin in basic_unreal_project_realpath.code_templates.items():
        assert isinstance(_name, str)
        assert isinstance(_plugin, code.CodeTemplate)


@pytest.mark.parametrize(
    "category,platform,expected",
    [
        (None, None, 0),
        ("Engine", None, 1),
        ("Engine", "Windows", 2),
    ],
)
def test_unreal_project_config_files(
    basic_unreal_project_realpath: project.UnrealProject,
    category: Optional[str],
    platform: Optional[str],
    expected: int,
) -> None:
    assert (
        len(list(basic_unreal_project_realpath.config_files(category, platform)))
        == expected
    )


def test_unreal_project_config(
    basic_unreal_project_realpath: project.UnrealProject,
    empty_file_unreal_project: project.UnrealProject,
    monkeypatch: Any,
) -> None:
    monkeypatch.setattr(project, "entry_points", lambda: {})
    assert isinstance(basic_unreal_project_realpath.config(), config.UnrealConfigParser)
    assert isinstance(empty_file_unreal_project.config(), config.UnrealConfigParser)


def test_unreal_project_engine(
    basic_unreal_project_realpath: project.UnrealProject,
    null_engine_unreal_project: project.UnrealProject,
    monkeypatch: Any,
) -> None:
    monkeypatch.setattr(project, "entry_points", lambda: {})
    assert isinstance(basic_unreal_project_realpath.engine, engine.UnrealEngine)
    assert null_engine_unreal_project.engine is None


def test_unreal_project_list_project_code_templates(
    basic_unreal_project_realpath: project.UnrealProject,
) -> None:
    assert (
        len(
            list(
                project.UnrealProject.list_project_code_templates(
                    basic_unreal_project_realpath
                )
            )
        )
        == 0
    )
